from .mainClass import Distance
from .vectorDistance  import Euclidean
#from .tools     import *

class DynamicTimeWarping(Distance):
	def __init__(self)-> None:
		"""
		Compute the Dynamic Time Warping (DTW) distance between two time series.
    
		DTW is a measure of similarity between two temporal sequences that may vary in speed.
		This class allows the computation of the DTW distance and the optimal alignment path between the sequences.
    
		Attributes:
		series_a (list or array): First time series.
		series_b (list or array): Second time series.
		distance_matrix (2D list): The accumulated cost matrix used to compute the DTW distance.
		dtw_distance (float): The computed DTW distance between series_a and series_b.
		"""
		super().__init__()
		self.type='vec_float'

	def compute(self, series_a, series_b):
		"""
		Compute the DTW distance between the two time series.
        
		Returns:
			float: The DTW distance.
		"""
		self.series_a = series_a
		self.series_b = series_b
		self.distance_matrix = None
		self.dtw_distance = None
        
		n = len(self.series_a)
		m = len(self.series_b)
		self.distance_matrix = [[float('inf')] * m for _ in range(n)]
		self.distance_matrix[0][0] = 0
        
		for i in range(1, n):
			for j in range(1, m):
				cost = abs(self.series_a[i] - self.series_b[j])
				self.distance_matrix[i][j] = cost + min(self.distance_matrix[i-1][j],    # Insertion
						self.distance_matrix[i][j-1],    # Deletion
						self.distance_matrix[i-1][j-1])  # Match

		self.dtw_distance = self.distance_matrix[-1][-1]
		return self.dtw_distance

	def get_optimal_path(self):
		"""
		Retrieve the optimal path that aligns the two time series with the minimum cost.
        
		Returns:
			list of tuples: The optimal path as a list of index pairs (i, j).
		"""
		i, j = len(self.series_a) - 1, len(self.series_b) - 1
		path = [(i, j)]
        
		while i > 0 and j > 0:
			if i == 0:
				j -= 1
			elif j == 0:
				i -= 1
			else:
				step = min(self.distance_matrix[i-1][j], self.distance_matrix[i][j-1], self.distance_matrix[i-1][j-1])
                
				if step == self.distance_matrix[i-1][j-1]:
					i -= 1
					j -= 1
				elif step == self.distance_matrix[i-1][j]:
					i -= 1
				else:
					j -= 1
            
				path.append((i, j))
        
		path.reverse()
		return path

from typing import List

#class dev pour les textes !!!!
class LongestCommonSubsequence(Distance):

	def __init__(self) -> None:
		super().__init__()
		self.type='vec_str'

		"""
		A class to compute the Longest Common Subsequence (LCS) between two text files.
		"""

	def _lcs_matrix(self, text1: str, text2: str) -> List[List[int]]:
		"""
		Constructs the LCS matrix for two input texts.
        
		:param text1: The first text as a string.
		:param text2: The second text as a string.
		:return: A 2D list (matrix) containing the lengths of LCS for substrings of text1 and text2.
		"""
		len1: int = len(text1)
		len2: int = len(text2)
        
		# Create a 2D matrix initialized with 0
		lcs_matrix: List[List[int]] = [[0] * (len2 + 1) for _ in range(len1 + 1)]

		for i in range(1, len1 + 1):
			for j in range(1, len2 + 1):
				if text1[i - 1] == text2[j - 1]:
					lcs_matrix[i][j] = lcs_matrix[i - 1][j - 1] + 1
				else:
					lcs_matrix[i][j] = max(lcs_matrix[i - 1][j], lcs_matrix[i][j - 1])

		return lcs_matrix

	def _backtrack_lcs(self, lcs_matrix: List[List[int]], text1: str, text2: str) -> str:
		"""
		Backtracks through the LCS matrix to reconstruct the longest common subsequence.
        
		:param lcs_matrix: A 2D list (matrix) containing the lengths of LCS for substrings of text1 and text2.
		:param text1: The first text as a string.
		:param text2: The second text as a string.
		:return: The longest common subsequence as a string.
		"""
		i: int = len(text1)
		j: int = len(text2)
		lcs: List[str] = []

		while i > 0 and j > 0:
			if text1[i - 1] == text2[j - 1]:
				lcs.append(text1[i - 1])
				i -= 1
				j -= 1
			elif lcs_matrix[i - 1][j] >= lcs_matrix[i][j - 1]:
				i -= 1
			else:
				j -= 1

		return ''.join(reversed(lcs))

	def compute(self, text1: str, text2: str) -> str:
		"""
		Computes the Longest Common Subsequence (LCS) between two texts.
        
		:param text1: The first text as a string.
		:param text2: The second text as a string.
		:return: The longest common subsequence as a string.
		"""
		# Compute the LCS matrix
		lcs_matrix: List[List[int]] = self._lcs_matrix(text1, text2)

		# Backtrack to find the actual LCS
		lcs: str = self._backtrack_lcs(lcs_matrix, text1, text2)

		return lcs
		
	def example(self):
		self.obj1_exemple = "AGGTAB"
		self.obj2_exemple = "GXTXAYB"
		sequence=self.compute(self.obj1_exemple,self.obj2_exemple)
		
		print(f"{self.__class__.__name__} distance between {self.obj1_exemple} and {self.obj2_exemple} is {sequence}")

class Frechet(Distance):

	def __init__(self)-> None:
		"""
		Initialize the FrechetDistance with two curves.

		:param curve_a: First curve, a list of tuples representing points (e.g., [(x1, y1), (x2, y2), ...])
		:param curve_b: Second curve, a list of tuples representing points (e.g., [(x1, y1), (x2, y2), ...])
		"""
		super().__init__()
		self.type='vec_tuple_float'


	def _c(self, i, j):
		"""
		Internal method to compute the discrete Fréchet distance using dynamic programming.

		:param i: Index in curve_a
		:param j: Index in curve_b
		:return: Fréchet distance between curve_a[0..i] and curve_b[0..j]
		"""
		if self.ca[i][j] > -1:
			return self.ca[i][j]
		elif i == 0 and j == 0:
			self.ca[i][j] = Euclidean().calculate(self.curve_a[0], self.curve_b[0])
		elif i > 0 and j == 0:
			self.ca[i][j] = max(self._c(i - 1, 0), Euclidean().calculate(self.curve_a[i], self.curve_b[0]))
		elif i == 0 and j > 0:
			self.ca[i][j] = max(self._c(0, j - 1), Euclidean().calculate(self.curve_a[0], self.curve_b[j]))
		elif i > 0 and j > 0:
			self.ca[i][j] = max(
				min(self._c(i - 1, j), self._c(i - 1, j - 1), self._c(i, j - 1)),
				Euclidean().calculate(self.curve_a[i], self.curve_b[j])
				)
		else:
			self.ca[i][j] = float('inf')
		return self.ca[i][j]

	def compute(self, curve_a, curve_b):
		"""
		Compute the Fréchet distance between the two curves.

		:return: The Fréchet distance between curve_a and curve_b
		"""
		self.curve_a = curve_a
		self.curve_b = curve_b
		self.ca = [[-1 for _ in range(len(curve_b))] for _ in range(len(curve_a))]
        
		return self._c(len(self.curve_a) - 1, len(self.curve_b) - 1)
	
	def example(self):
		self.obj1_example = [(0, 0), (1, 1), (2, 2)]
		self.obj2_example = [(0, 0), (1, 2), (2, 3)]
		distance=self.compute(self.obj1_example,self.obj2_example)
		print(f"{self.__class__.__name__} distance between {self.obj1_example} and {self.obj2_example} is {distance:.2f}")







from .mainClass import *
from .tools     import Vector,Matrix

class Mahalanobis(Distance):
	
	def __init__(self)-> None:
		super().__init__()
		self.type='list_list'

	def mean(self,data:list[list]):
		"""
		Calculate the mean of each dimension in the dataset.
    
		:param data: A dataset as a list of points (list of lists)
		:return: Mean of each dimension as a list
		"""
		n = len(data)
		d = len(data[0])
		mean = [0] * d
    
		for point in data:
			for i in range(d):
				mean[i] += point[i]
    
		mean = [x / n for x in mean]
		return mean

	def covariance_matrix(self,data, mean):
		"""
		Calculate the covariance matrix of the dataset.
    
		:param data: A dataset as a list of points (list of lists)
		:param mean: Mean of each dimension as a list
		:return: Covariance matrix as a list of lists
		"""
		n = len(data)
		d = len(data[0])
		cov_matrix = [[0] * d for _ in range(d)]
    
		for point in data:
			diff = [point[i] - mean[i] for i in range(d)]
			for i in range(d):
				for j in range(d):
					cov_matrix[i][j] += diff[i] * diff[j]
    
		cov_matrix = [[x / (n - 1) for x in row] for row in cov_matrix]
		return cov_matrix

	def matrix_inverse(self,matrix):
		"""
		Calculate the inverse of a matrix using Gauss-Jordan elimination.
    
		:param matrix: A square matrix as a list of lists
		:return: Inverse of the matrix as a list of lists
		"""
		n = len(matrix)
		identity = [[float(i == j) for i in range(n)] for j in range(n)]
		augmented = [row + identity_row for row, identity_row in zip(matrix, identity)]
		for i in range(n):
			pivot = augmented[i][i]
			for j in range(2 * n):
				augmented[i][j] /= pivot
			for k in range(n):
				if k != i:
					factor = augmented[k][i]
					for j in range(2 * n):
						augmented[k][j] -= factor * augmented[i][j]
    
		inverse = [row[n:] for row in augmented]
		return inverse

	def compute(self,point :list, data :list[list]) -> float:
		"""
		Calculate the Mahalanobis distance between a point and a dataset.
    
		:param point: A point as a list of coordinates
		:param data: A dataset as a list of points (list of lists)
		:return: Mahalanobis distance between the point and the dataset
		:raises ValueError: If the point dimensions do not match the dataset dimensions
		! lever une execption si la matrice est singulière
		"""
    
		mean_data = self.mean(data)
		cov_matrix = self.covariance_matrix(data, mean_data)
		cov_matrix_inv = self.matrix_inverse(cov_matrix)
    
		diff = [point[i] - mean_data[i] for i in range(len(point))]
    
		# Matrix multiplication: diff^T * cov_matrix_inv * diff
		result = 0
		for i in range(len(diff)):
			for j in range(len(diff)):
				result += diff[i] * cov_matrix_inv[i][j] * diff[j]
    
		return result**0.5
		
	def example(self):
		self.obj1_exemple=Vector.vec_float_1
		self.obj2_exemple = [
    [2, 1, 0],
    [2, 3, 4],
    [3, 4, 5],
    [4, 5, 6],
]
		super().example()
		




class MahalanobisTaguchi(Distance):
	
    def __init__(self, reference_group :list[list])-> None:
        """
        Initialize the MahalanobisTaguchi class with a reference group.

        :param reference_group: A list of lists where each inner list is a data point in the reference group.
        """
        super().__init__()
        self.type='vec_float'


        self.reference_group = reference_group
        self.mean_vector = self.calculate_mean_vector()
        self.covariance_matrix = Matrix.covariance_matrix()
        self.inverse_covariance_matrix = Matrix.invert_matrix(self.covariance_matrix)

    def calculate_mean_vector(self):
        """
        Calculate the mean vector of the reference group.

        :return: A list representing the mean vector.
        """
        num_points = len(self.reference_group)
        num_dimensions = len(self.reference_group[0])

        mean_vector = [0] * num_dimensions

        for point in self.reference_group:
            for i in range(num_dimensions):
                mean_vector[i] += point[i]

        mean_vector = [x / num_points for x in mean_vector]
        return mean_vector

    def compute(self, data_point)-> float:
        """
        Calculate the Mahalanobis-Taguchi distance for a given data point.

        :param data_point: A list representing the data point to be evaluated.
        :return: The Mahalanobis-Taguchi distance as a float.
        """
        diff_vector = [data_point[i] - self.mean_vector[i] for i in range(len(self.mean_vector))]
        
        # Matrix multiplication with the inverse covariance matrix
        temp_vector = [0] * len(diff_vector)
        for i in range(len(diff_vector)):
            for j in range(len(diff_vector)):
                temp_vector[i] += diff_vector[j] * self.inverse_covariance_matrix[j][i]

        # Final dot product to get the Mahalanobis-Taguchi distance
        distance_squared = sum(temp_vector[i] * diff_vector[i] for i in range(len(diff_vector)))
        return distance_squared ** 0.5
        
    def example(self):
      self.obj1_exemple=Vector.vec_float_1
      self.obj2_exemple = [
    [2, 1, 0],
    [2, 3, 4],
    [3, 4, 5],
    [4, 5, 6],
]
    super().example()





from .mainClass import *
from .vectorDistance import Euclidean

from typing import List, Dict, Set, Tuple, TypeVar, Callable

#claude ai
T = TypeVar('T')
class Levenshtein(Distance):
	def __init__(self, insert_cost: float = 1.0, delete_cost: float = 1.0, replace_cost: float = 1.0):
		super().__init__()
		self.type='text'

		self.insert_cost = insert_cost
		self.delete_cost = delete_cost
		self.replace_cost = replace_cost
		self.equality_func = lambda x, y: x == y


	def compute(self, s1: List[T], s2: List[T]) -> float:
		m: int = len(s1)
		n: int = len(s2)
        
		# Initialisation de la matrice
		dp: List[List[float]] = [[0.0 for _ in range(n + 1)] for _ in range(m + 1)]
        
		# Remplissage de la première ligne et de la première colonne
		for i in range(m + 1):
			dp[i][0] = i * self.delete_cost
		for j in range(n + 1):
			dp[0][j] = j * self.insert_cost
        
		# Calcul de la distance
		for i in range(1, m + 1):
			for j in range(1, n + 1):
				if self.equality_func(s1[i-1], s2[j-1]):
					dp[i][j] = dp[i-1][j-1]
				else:
					dp[i][j] = min(
						dp[i-1][j] + self.delete_cost,  # Suppression
						dp[i][j-1] + self.insert_cost,  # Insertion
						dp[i-1][j-1] + self.replace_cost  # Remplacement
						)
        
		return dp[m][n]

	@staticmethod
	def levenshtein_distance_words(sentence1: str, sentence2: str) -> float:
			words1 = sentence1.split()
			words2 = sentence2.split()
			lev = Levenshtein()
			return lev.compute(words1, words2)



'''
class Levenshtein(Distance):
	
	def __init__(self)-> None:
		super().__init__()
		
	def compute(self,s1: str, s2: str) -> int:
		
		dp = [[0] * (len(s2) + 1) for _ in range(len(s1) + 1)]

		for i in range(len(s1) + 1):
			dp[i][0] = i
		for j in range(len(s2) + 1):
			dp[0][j] = j

		for i in range(1, len(s1) + 1):
			for j in range(1, len(s2) + 1):
				if s1[i - 1] == s2[j - 1]:
					cost = 0
				else:
					cost = 1
				dp[i][j] = min(dp[i - 1][j] + 1,        # Suppression
                           dp[i][j - 1] + 1,        # Insertion
                           dp[i - 1][j - 1] + cost) # Substitution

		return dp[len(s1)][len(s2)]
		
	def exemple(self):
		self.obj1_exemple = "kitten"
		self.obj2_exemple = "sitting"
		super().exemple()
'''
class DamerauLevenshtein(Distance):
	
	def __init__(self)-> None:
		super().__init__()
		self.type='str'

	def compute(self,s1 :str, s2 :str) -> int:
		d = {}
		lenstr1 = len(s1)
		lenstr2 = len(s2)

		for i in range(-1, lenstr1 + 1):
			d[(i, -1)] = i + 1
		for j in range(-1, lenstr2 + 1):
			d[(-1, j)] = j + 1

		for i in range(lenstr1):
			for j in range(lenstr2):
				cost = 0 if s1[i] == s2[j] else 1
				d[(i, j)] = min(
					d[(i - 1, j)] + 1,  # suppresion
					d[(i, j - 1)] + 1,  # insertion
					d[(i - 1, j - 1)] + cost,  # substitution
				)
				if i > 0 and j > 0 and s1[i] == s2[j - 1] and s1[i - 1] == s2[j]:
					d[(i, j)] = min(d[(i, j)], d[(i - 2, j - 2)] + cost)  # transposition

		return d[lenstr1 - 1, lenstr2 - 1]
		
	def exemple(self):
		self.obj1_exemple = "ca"
		self.obj2_exemple = "abc"

		super().exemple()

class Hamming(Distance):
	
	def __init__(self)-> None:
		super().__init__()
		self.type='vec_int'

	def compute(self,str1: list, str2: list) -> int:
		"""
		Calculate the Hamming distance between two strings.
    
		:param str1: First string
		:param str2: Second string
		:return: Hamming distance between str1 and str2
		:raises ValueError: If the strings are not of the same length
		"""
		return sum(el1 != el2 for el1, el2 in zip(str1, str2))


#claude ai
from typing import List, Union, Dict

class Cosine(Distance):
	def __init__(self)-> None:
		super().__init__()
		#self.type='vec_float'
		self.type='text'
		
	@staticmethod
	def string_to_vector(text: str) -> Dict[str, int]:
		"""Converts a string into a word frequency vector."""
		words: List[str] = text.lower().split()
		return {word: words.count(word) for word in set(words)}

	@staticmethod
	def dot_product(v1: Dict[str, Union[int, float]], v2: Dict[str, Union[int, float]]) -> float:
		"""Calculates the dot product of two vectors."""
		return sum(v1.get(key, 0) * v2.get(key, 0) for key in set(v1) | set(v2))

	@staticmethod
	def magnitude(vector: Dict[str, Union[int, float]]) -> float:
		"""Calculates the magnitude of a vector."""
		return sum(value ** 2 for value in vector.values())**0.5

	@classmethod
	def cosine_similarity(cls, v1: Dict[str, Union[int, float]], v2: Dict[str, Union[int, float]]) -> float:
		"""Calculates the cosine similarity between two vectors."""
		dot_prod: float = cls.dot_product(v1, v2)
		mag1: float = cls.magnitude(v1)
		mag2: float = cls.magnitude(v2)
        
		if mag1 == 0 or mag2 == 0:
			return 0.0
        
		return dot_prod / (mag1 * mag2)

	@classmethod
	def compute(cls, v1: Union[str, List[float]], v2: Union[str, List[float]]) -> float:
		"""Calculates the cosine distance between two vectors or strings."""
		if isinstance(v1, str) and isinstance(v2, str):
			v1_dict: Dict[str, int] = cls.string_to_vector(v1)
			v2_dict: Dict[str, int] = cls.string_to_vector(v2)
		elif isinstance(v1, list) and isinstance(v2, list):
			v1_dict: Dict[str, float] = {str(i): val for i, val in enumerate(v1)}
			v2_dict: Dict[str, float] = {str(i): val for i, val in enumerate(v2)}
		else:
			raise ValueError("Inputs must be either two strings or two lists of floats.")
        
		similarity: float = cls.cosine_similarity(v1_dict, v2_dict)
		return 1.0 - similarity

class CosineInverse(Distance):
	
	def __init__(self)-> None:
		super().__init__()
		self.type='text'
		
	def compute(self,vec1 :list, vec2 :list)-> float:
		return 1-Cosine().compute(vec1,vec2)
		
class CosineTF(Distance):

    def __init__(self)-> None:
      super().__init__()
      self.type='text'
    """
    A class to compute the Cosine Similarity between two text documents based on term frequency vectors.
    """


    def _tokenize(self, document: str) -> List[str]:
        """
        Tokenizes the document into individual words (tokens).
        
        :param document: The input document as a string.
        :return: A list of tokens (words) from the document.
        """
        return document.lower().split()

    def _compute_term_frequencies(self, tokens: List[str]) -> Dict[str, int]:
        """
        Computes the term frequency (TF) of each token in the document.
        
        :param tokens: A list of tokens extracted from the document.
        :return: A dictionary where keys are tokens and values are their frequencies in the document.
        """
        term_frequencies: Dict[str, int] = {}
        for token in tokens:
            if token in term_frequencies:
                term_frequencies[token] += 1
            else:
                term_frequencies[token] = 1
        return term_frequencies

    def _dot_product(self, vec1: Dict[str, int], vec2: Dict[str, int]) -> float:
        """
        Computes the dot product of two term frequency vectors.
        
        :param vec1: The first term frequency vector as a dictionary.
        :param vec2: The second term frequency vector as a dictionary.
        :return: The dot product of the two vectors.
        """
        dot_product: float = 0.0
        for term, freq in vec1.items():
            if term in vec2:
                dot_product += freq * vec2[term]
        return dot_product

    def _magnitude(self, vec: Dict[str, int]) -> float:
        """
        Computes the magnitude (Euclidean norm) of a term frequency vector.
        
        :param vec: A term frequency vector as a dictionary.
        :return: The magnitude of the vector.
        """
        magnitude: float = math.sqrt(sum(freq ** 2 for freq in vec.values()))
        return magnitude

    def compute(self, document1: str, document2: str) -> float:
        """
        Computes the Cosine Similarity between two documents.
        
        :param document1: The first document as a string.
        :param document2: The second document as a string.
        :return: The cosine similarity score between 0 and 1.
        """
        # Tokenize both documents
        tokens1: List[str] = self._tokenize(document1)
        tokens2: List[str] = self._tokenize(document2)

        # Compute term frequencies
        tf1: Dict[str, int] = self._compute_term_frequencies(tokens1)
        tf2: Dict[str, int] = self._compute_term_frequencies(tokens2)

        # Compute the dot product of the two vectors
        dot_product: float = self._dot_product(tf1, tf2)

        # Compute the magnitude of each vector
        magnitude1: float = self._magnitude(tf1)
        magnitude2: float = self._magnitude(tf2)

        # Avoid division by zero
        if magnitude1 == 0 or magnitude2 == 0:
            return 0.0

        # Compute and return the cosine similarity
        cosine_similarity: float = dot_product / (magnitude1 * magnitude2)
        return cosine_similarity
'''    
import math
from collections import Counter

class TFIDFDistance(Distance):

    def __init__(self, corpus: List[str]) -> None:
        """
        Initialisation de la classe TFIDFDistance.
        
        :param corpus: Liste de documents (chaîne de caractères) utilisés pour calculer les fréquences globales des termes.
        """
        super().__init__()
        self.type='text'
        self.corpus = corpus
        self.term_frequencies = self._compute_term_frequencies(corpus)
        self.document_frequencies = self._compute_document_frequencies(corpus)

    def _compute_term_frequencies(self, corpus: List[str]) -> List[Dict[str, int]]:
        """
        Calcule la fréquence des termes pour chaque document du corpus.
        
        :param corpus: Liste de documents.
        :return: Liste de dictionnaires de fréquences de termes pour chaque document.
        """
        term_frequencies = [Counter(doc.split()) for doc in corpus]
        return term_frequencies

    def _compute_document_frequencies(self, corpus: List[str]) -> Dict[str, int]:
        """
        Calcule la fréquence inverse des documents (DF) pour chaque terme du corpus.
        
        :param corpus: Liste de documents.
        :return: Dictionnaire de fréquence inverse des documents pour chaque terme.
        """
        doc_freq: Dict[str, int] = {}
        for document in corpus:
            unique_terms = set(document.split())
            for term in unique_terms:
                doc_freq[term] = doc_freq.get(term, 0) + 1
        return doc_freq

    def _compute_tf_idf(self, document: str) -> Dict[str, float]:
        """
        Calcule les valeurs TF-IDF pour un document donné.
        
        :param document: Chaîne de caractères représentant un document.
        :return: Dictionnaire des valeurs TF-IDF pour chaque terme du document.
        """
        term_freq: Dict[str, int] = Counter(document.split())
        total_terms = len(document.split())
        tf_idf: Dict[str, float] = {}
        
        for term, freq in term_freq.items():
            tf = freq / total_terms
            idf = math.log(len(self.corpus) / (1 + self.document_frequencies.get(term, 0)))
            tf_idf[term] = tf * idf

        return tf_idf

    def _cosine_similarity(self, vec1: Dict[str, float], vec2: Dict[str, float]) -> float:
        """
        Calcule la similarité cosinus entre deux vecteurs TF-IDF.
        
        :param vec1: Premier vecteur TF-IDF.
        :param vec2: Deuxième vecteur TF-IDF.
        :return: Valeur de similarité cosinus.
        """
        intersection = set(vec1.keys()).intersection(vec2.keys())
        dot_product = sum(vec1[term] * vec2[term] for term in intersection)
        
        norm1 = math.sqrt(sum([value ** 2 for value in vec1.values()]))
        norm2 = math.sqrt(sum([value ** 2 for value in vec2.values()]))
        
        if not norm1 or not norm2:
            return 0.0
        
        return dot_product / (norm1 * norm2)

    def compare(self, text1: str, text2: str) -> float:
        """
        Compare deux textes en utilisant la distance TF-IDF (Cosine Similarity).
        
        :param text1: Premier document à comparer.
        :param text2: Deuxième document à comparer.
        :return: Valeur de la similarité entre 0 et 1 (plus elle est proche de 1, plus les documents sont similaires).
        """
        tf_idf1 = self._compute_tf_idf(text1)
        tf_idf2 = self._compute_tf_idf(text2)
        return self._cosine_similarity(tf_idf1, tf_idf2)
'''


class TFIDF(Distance):

    def __init__(self)-> None:
      super().__init__()
      self.type='text'
    """
    A class to compute the similarity between two documents based on TF-IDF (Term Frequency-Inverse Document Frequency).
    """

    def _tokenize(self, document: str) -> List[str]:
        """
        Tokenizes the document into individual words (tokens).
        
        :param document: The input document as a string.
        :return: A list of tokens (words) from the document.
        """
        return document.lower().split()

    def _compute_term_frequencies(self, tokens: List[str]) -> Dict[str, int]:
        """
        Computes the term frequency (TF) of each token in the document.
        
        :param tokens: A list of tokens extracted from the document.
        :return: A dictionary where keys are tokens and values are their frequencies in the document.
        """
        term_frequencies: Dict[str, int] = {}
        for token in tokens:
            if token in term_frequencies:
                term_frequencies[token] += 1
            else:
                term_frequencies[token] = 1
        return term_frequencies

    def _compute_document_frequencies(self, documents: List[List[str]]) -> Dict[str, int]:
        """
        Computes the document frequency (DF) for each token across multiple documents.
        
        :param documents: A list of tokenized documents (each document is a list of tokens).
        :return: A dictionary where keys are tokens and values are the number of documents containing the token.
        """
        document_frequencies: Dict[str, int] = {}
        for tokens in documents:
            unique_tokens: Set[str] = set(tokens)
            for token in unique_tokens:
                if token in document_frequencies:
                    document_frequencies[token] += 1
                else:
                    document_frequencies[token] = 1
        return document_frequencies

    def _compute_tfidf(self, term_freq: Dict[str, int], doc_freq: Dict[str, int], total_docs: int) -> Dict[str, float]:
        """
        Computes the TF-IDF value for each token in the document.
        
        :param term_freq: Term frequency for a document.
        :param doc_freq: Document frequency for all documents.
        :param total_docs: Total number of documents.
        :return: A dictionary where keys are tokens and values are their TF-IDF values.
        """
        tfidf: Dict[str, float] = {}
        for term, freq in term_freq.items():
            tf: float = freq / sum(term_freq.values())
            idf: float = math.log(total_docs / (1 + doc_freq.get(term, 0)))
            tfidf[term] = tf * idf
        return tfidf

    def _dot_product(self, vec1: Dict[str, float], vec2: Dict[str, float]) -> float:
        """
        Computes the dot product of two TF-IDF vectors.
        
        :param vec1: The first TF-IDF vector as a dictionary.
        :param vec2: The second TF-IDF vector as a dictionary.
        :return: The dot product of the two vectors.
        """
        dot_product: float = 0.0
        for term, value in vec1.items():
            if term in vec2:
                dot_product += value * vec2[term]
        return dot_product

    def _magnitude(self, vec: Dict[str, float]) -> float:
        """
        Computes the magnitude (Euclidean norm) of a TF-IDF vector.
        
        :param vec: A TF-IDF vector as a dictionary.
        :return: The magnitude of the vector.
        """
        magnitude: float = math.sqrt(sum(value ** 2 for value in vec.values()))
        return magnitude

    def compute(self, document1: str, document2: str, corpus: List[str]) -> float:
        """
        Computes the TF-IDF similarity (cosine similarity) between two documents based on a corpus of documents.
        
        :param document1: The first document as a string.
        :param document2: The second document as a string.
        :param corpus: A list of documents representing the corpus.
        :return: The TF-IDF similarity score between 0 and 1.
        """
        # Tokenize both documents and the corpus
        tokens1: List[str] = self._tokenize(document1)
        tokens2: List[str] = self._tokenize(document2)
        tokenized_corpus: List[List[str]] = [self._tokenize(doc) for doc in corpus]

        # Add the documents to the corpus for frequency calculations
        tokenized_corpus.append(tokens1)
        tokenized_corpus.append(tokens2)

        # Compute term frequencies for both documents
        tf1: Dict[str, int] = self._compute_term_frequencies(tokens1)
        tf2: Dict[str, int] = self._compute_term_frequencies(tokens2)

        # Compute document frequencies from the corpus
        doc_freq: Dict[str, int] = self._compute_document_frequencies(tokenized_corpus)

        # Total number of documents
        total_docs: int = len(tokenized_corpus)

        # Compute TF-IDF vectors for both documents
        tfidf1: Dict[str, float] = self._compute_tfidf(tf1, doc_freq, total_docs)
        tfidf2: Dict[str, float] = self._compute_tfidf(tf2, doc_freq, total_docs)

        # Compute the dot product of the two TF-IDF vectors
        dot_product: float = self._dot_product(tfidf1, tfidf2)

        # Compute the magnitude of each TF-IDF vector
        magnitude1: float = self._magnitude(tfidf1)
        magnitude2: float = self._magnitude(tfidf2)

        # Avoid division by zero
        if magnitude1 == 0 or magnitude2 == 0:
            return 0.0

        # Compute and return the TF-IDF cosine similarity
        tfidf_similarity: float = dot_product / (magnitude1 * magnitude2)
        return tfidf_similarity




class SimHash(Distance):

    def __init__(self)-> None:
      super().__init__()
      self.type='text'
    """
    A class to compute the SimHash of a document and to compare the SimHash values of two documents
    to measure their similarity.
    """

    def _hash(self, token: str) -> int:
        """
        Converts a token (string) into a 64-bit integer hash.

        :param token: The input string (token) to be hashed.
        :return: A 64-bit integer hash of the input token.
        """
        return int.from_bytes(token.encode('utf-8'), 'little') % (1 << 64)

    def _tokenize(self, document: str) -> List[str]:
        """
        Tokenizes the document into individual words (or tokens).

        :param document: The input document as a string.
        :return: A list of tokens (words) from the document.
        """
        return document.split()

    def _compute_weights(self, tokens: List[str]) -> Dict[str, int]:
        """
        Computes the weight of each token based on its frequency in the document.

        :param tokens: A list of tokens extracted from the document.
        :return: A dictionary where keys are tokens and values are their frequencies (weights).
        """
        token_weights: Dict[str, int] = {}
        for token in tokens:
            if token in token_weights:
                token_weights[token] += 1
            else:
                token_weights[token] = 1
        return token_weights

    def _compute_simhash(self, token_weights: Dict[str, int]) -> int:
        """
        Computes the SimHash value from the token weights using a weighted hash for each token.

        :param token_weights: A dictionary of token weights.
        :return: A 64-bit integer representing the SimHash value.
        """
        bit_vector: List[int] = [0] * 64

        for token, weight in token_weights.items():
            token_hash: int = self._hash(token)

            for i in range(64):
                # Extract each bit from the 64-bit hash
                bit = (token_hash >> i) & 1
                # Adjust the bit vector based on the weight
                if bit == 1:
                    bit_vector[i] += weight
                else:
                    bit_vector[i] -= weight

        # Convert bit_vector into a 64-bit SimHash
        simhash_value: int = 0
        for i in range(64):
            if bit_vector[i] > 0:
                simhash_value |= (1 << i)

        return simhash_value

    def compute(self, document: str) -> int:
        """
        Computes the SimHash value of a given document.

        :param document: The input document as a string.
        :return: A 64-bit integer representing the SimHash value.
        """
        tokens: List[str] = self._tokenize(document)
        token_weights: Dict[str, int] = self._compute_weights(tokens)
        simhash_value: int = self._compute_simhash(token_weights)
        return simhash_value

#claude ai
from typing import List, Union

class Hamming(Distance):

    def __init__(self)-> None:
      super().__init__()
      self.type='text'
      
    @staticmethod
    def compute(v1: Union[str, List[int]], v2: Union[str, List[int]]) -> int:
        """
        Calculate the Hamming distance between two strings or two lists of integers.
        
        Args:
            v1: First string or list of integers
            v2: Second string or list of integers
        
        Returns:
            The Hamming distance between v1 and v2
        
        Raises:
            ValueError: If inputs are not of the same type or have different lengths
        """
        if isinstance(v1, str) and isinstance(v2, str):
            if len(v1) != len(v2):
                raise ValueError("Strings must have the same length")
            return sum(c1 != c2 for c1, c2 in zip(v1, v2))
        
        elif isinstance(v1, list) and isinstance(v2, list):
            if len(v1) != len(v2):
                raise ValueError("Lists must have the same length")
            return sum(i1 != i2 for i1, i2 in zip(v1, v2))
        
        else:
            raise ValueError("Inputs must be either two strings or two lists of integers")

    @classmethod
    def normalized_distance(cls, v1: Union[str, List[int]], v2: Union[str, List[int]]) -> float:
        """
        Calculate the normalized Hamming distance between two strings or two lists of integers.
        
        The normalized distance is the Hamming distance divided by the length of the inputs.
        This gives a value between 0 and 1, where 0 means the inputs are identical,
        and 1 means they are completely different.
        
        Args:
            v1: First string or list of integers
            v2: Second string or list of integers
        
        Returns:
            The normalized Hamming distance between v1 and v2
        """
        distance = cls.compute(v1, v2)
        return distance / len(v1)  # len(v1) is safe to use as we've checked they're equal in distance()

    def similarity(self, document1: str, document2: str) -> float:
        """
        Computes the similarity between two documents based on their SimHash values.

        :param document1: The first document as a string.
        :param document2: The second document as a string.
        :return: A similarity score between 0 (completely different) and 1 (identical).
        """
        simhash1: int = self.compute(document1)
        simhash2: int = self.compute(document2)

        # Calculate Hamming distance
        hamming_dist: int = self.hamming_distance(simhash1, simhash2)
        max_bits: int = 64

        # Compute similarity as the fraction of matching bits
        similarity_score: float = (max_bits - hamming_dist) / max_bits
        return similarity_score



from typing import Optional
#from gensim.models import KeyedVectors
#from gensim.similarities import WmdSimilarity
#from gensim.corpora.dictionary import Dictionary

import gensim.downloader as api

# Import and download stopwords from NLTK.
from nltk.corpus import stopwords
from nltk import download

class WordMoversDistance(Distance):

    def __init__(self) -> None:
        """
        Initialize the WordMoversDistance class with a pre-trained word embedding model and a corpus of texts.

        :param model: A pre-trained word embedding model (e.g., Word2Vec or GloVe).
        :param texts: A list of text documents to be compared.
        """
        super().__init__()
        self.type='text'
        download('stopwords')  # Download stopwords list.
        self.stop_words = stopwords.words('english')
        self.model = api.load('word2vec-google-news-300')
        #or download -> https://www.kaggle.com/datasets/leadbest/googlenewsvectorsnegative300?resource=download


    def preprocess(self,sentence):
        return [w for w in sentence.lower().split() if w not in self.stop_words]
        

    def compute_distance(self, text1: str, text2: str) -> Optional[float]:
        """
        Compute the Word Mover's Distance between two text documents.

        :param text1: The first text document.
        :param text2: The second text document.
        :return: The Word Mover's Distance between the two documents, or None if it cannot be computed.
        """
        tokens1: List[str] = self.preprocess(text1)
        tokens2: List[str] = self.preprocess(text2)
                
        if tokens1 and tokens2:
            return self.model.wmdistance(tokens1, tokens2)
        return None

    def compute(self,text1: str, text2: str) -> Optional[float]:
        """
        Compare two text files using Word Mover's Distance and a pre-trained word embedding model.

        :param text1: first text.
        :param text2: second text.
        :return: The Word Mover's Distance between the two text files, or None if it cannot be computed.
        """
        try:
            # Load the pre-trained word embedding model
            #model: KeyedVectors = KeyedVectors.load_word2vec_format(model_path, binary=True)

            # Compute and return the Word Mover's Distance
            return self.model.wmdistance(text1, text2)

        except Exception as e:
            print(f"Error processing files: {e}")
            return None




from transformers import BertTokenizer, BertModel
import torch

class BERTBasedDistance(Distance):

    def __init__(self, model_name: str = "bert-base-uncased") -> None:
        """
        Initialize the BERTBasedDistance class with a BERT model and tokenizer.

        :param model_name: The name of the pre-trained BERT model to use.
        """
        super().__init__()
        self.type='text'
        self.tokenizer: BertTokenizer = BertTokenizer.from_pretrained(model_name)
        self.model: BertModel = BertModel.from_pretrained(model_name)

    def _encode_text(self, text: str) -> torch.Tensor:
        """
        Encode a text into a BERT embedding.

        :param text: The input text to encode.
        :return: A tensor representing the embedding of the input text.
        """
        inputs = self.tokenizer(text, return_tensors="pt", max_length=512, truncation=True, padding="max_length")
        with torch.no_grad():
            outputs = self.model(**inputs)
        return outputs.last_hidden_state.mean(dim=1)  # Mean pooling over the sequence length

    def compute_distance(self, text1: str, text2: str) -> float:
        """
        Compute the cosine similarity between the BERT embeddings of two texts.

        :param text1: The first text document.
        :param text2: The second text document.
        :return: The cosine similarity between the embeddings of the two documents.
        """
        embedding1: torch.Tensor = self._encode_text(text1)
        embedding2: torch.Tensor = self._encode_text(text2)

        # Cosine similarity between the two embeddings
        similarity: float = torch.nn.functional.cosine_similarity(embedding1, embedding2).item()
        return similarity

    @staticmethod
    def compute(text1: str, text2: str) -> Optional[float]:
        """
        Compare two text files using BERT embeddings to calculate semantic similarity.

        :param file1_path: Path to the first text file.
        :param file2_path: Path to the second text file.
        :return: The similarity score between the two text files based on BERT embeddings.
        """
        try:

            # Initialize BERTBasedDistance object
            bert_distance = BERTBasedDistance()

            # Compute and return the similarity score
            return bert_distance.compute_distance(text1, text2)

        except Exception as e:
            print(f"Error processing files: {e}")
            return None
            
class Jaro(Distance):
	
	def __init__(self)-> None:
		super().__init__()
		self.type='str'

	def compute(self,s1 :str, s2 :str) -> float:
		"""
		Calculate the Jaro similarity between two strings.
    
		:param s1: The first string
		:param s2: The second string
		:return: Jaro similarity between the two strings
		"""
		if s1 == s2:
			return 1.0

		len_s1 = len(s1)
		len_s2 = len(s2)

		if len_s1 == 0 or len_s2 == 0:
			return 0.0

		match_distance = max(len_s1, len_s2) // 2 - 1

		s1_matches = [False] * len_s1
		s2_matches = [False] * len_s2

		matches = 0
		transpositions = 0

		for i in range(len_s1):
			start = max(0, i - match_distance)
			end = min(i + match_distance + 1, len_s2)

			for j in range(start, end):
				if s2_matches[j]:
					continue
				if s1[i] != s2[j]:
					continue
				s1_matches[i] = True
				s2_matches[j] = True
				matches += 1
				break

		if matches == 0:
			return 0.0

		k = 0
		for i in range(len_s1):
			if not s1_matches[i]:
				continue
			while not s2_matches[k]:
				k += 1
			if s1[i] != s2[k]:
				transpositions += 1
			k += 1

		return (matches / len_s1 + matches / len_s2 + (matches - transpositions // 2) / matches) / 3.0
		
	def exemple(self):
		self.obj1_exemple = "martha"
		self.obj2_exemple = "marhta"
		super().exemple()
		
class JaroWinkler(Distance):
	
	def __init__(self)-> None:
		super().__init__()
		self.type='text'

	def compute(self,s1 :str, s2 :str, p=0.1) -> float:
		"""
		Calculate the Jaro-Winkler distance between two strings.
    
		:param s1: The first string
		:param s2: The second string
		:param p: The scaling factor, usually 0.1
		:return: Jaro-Winkler distance between the two strings
		"""
		jaro_sim = Jaro().calculate(s1, s2)

		prefix_length = 0
		max_prefix_length = 4

		for i in range(min(len(s1), len(s2))):
			if s1[i] == s2[i]:
				prefix_length += 1
			else:
				break
			if prefix_length == max_prefix_length:
				break

		jaro_winkler_sim = jaro_sim + (prefix_length * p * (1 - jaro_sim))
		return jaro_winkler_sim
		
	def exemple(self):
		self.obj1_exemple = "martha"
		self.obj2_exemple = "marhta"
		super().exemple()

class OverlapCoefficient(Distance):

    def __init__(self) -> None:
      super().__init__()
      self.type='set_text'

    def compute(self, set1: Set[str], set2: Set[str]) -> float:
        """
        Calcule le coefficient de chevauchement (Overlap Coefficient) entre deux ensembles de mots.
        
        :param set1: Premier ensemble de mots.
        :param set2: Deuxième ensemble de mots.
        :return: Coefficient de chevauchement entre les deux ensembles, entre 0 et 1.
        """
        if not set1 or not set2:
            return 0.0
        
        # Calcul de l'intersection des deux ensembles
        intersection_size: int = len(set1.intersection(set2))
        
        # Taille du plus petit ensemble
        min_size: int = min(len(set1), len(set2))
        
        # Coefficient de chevauchement
        overlap_coefficient: float = intersection_size / min_size
        
        return overlap_coefficient


class SorensenDice(Distance):
	
	def __init__(self)-> None:
		super().__init__()
		self.type='text'
		
	def compute(self,str1 :str, str2 :str) -> float:
		# Convert strings to sets of bigrams
		bigrams1 = {str1[i:i+2] for i in range(len(str1) - 1)}
		bigrams2 = {str2[i:i+2] for i in range(len(str2) - 1)}
    
		# Calculate the intersection and the sizes of the sets
		intersection = len(bigrams1 & bigrams2)
		size1 = len(bigrams1)
		size2 = len(bigrams2)
    
		# Calculate the Sørensen-Dice coefficient
		sorensen_dice_coeff = 2 * intersection / (size1 + size2)
    
		# The distance is 1 minus the coefficient
		distance = 1 - sorensen_dice_coeff
    
		return distance
		
	def exemple(self):
		self.obj1_exemple = "night"
		self.obj2_exemple = "nacht"

		super().exemple()
		
from collections import Counter
import math

class BagOfWordsDistance(Distance):

    def __init__(self) -> None:
      super().__init__()
      self.type='text'
      
    def compute(self, text1: str, text2: str) -> float:
        """
        Calcule la distance entre deux textes en utilisant la représentation Bag-of-Words.

        :param text1: Premier texte.
        :param text2: Deuxième texte.
        :return: Distance entre les deux textes, entre 0 et 1.
        """
        # Création des sacs de mots pour les deux textes
        bow1: Dict[str, int] = self._text_to_bow(text1)
        bow2: Dict[str, int] = self._text_to_bow(text2)

        # Union de tous les mots dans les deux textes
        all_words: List[str] = list(set(bow1.keys()).union(set(bow2.keys())))

        # Vecteurs de fréquences des mots pour chaque texte
        vec1: List[int] = [bow1.get(word, 0) for word in all_words]
        vec2: List[int] = [bow2.get(word, 0) for word in all_words]

        # Calcul de la distance entre les deux vecteurs (utilisation de la distance euclidienne ici)
        distance: float = Euclidean().compute(vec1, vec2)
        
        return distance

    def _text_to_bow(self, text: str) -> Dict[str, int]:
        """
        Convertit un texte en un sac de mots (Bag of Words).

        :param text: Texte à convertir.
        :return: Dictionnaire représentant la fréquence de chaque mot dans le texte.
        """
        # Découper le texte en mots
        words: List[str] = text.lower().split()
        
        # Créer un sac de mots avec les fréquences de chaque mot
        bow: Dict[str, int] = dict(Counter(words))
        
        return bow
        
from gensim.models import FastText

class FastTextDistance(Distance):

    def __init__(self, model: FastText) -> None:
        """
        Initialisation de la classe FastTextDistance.

        :param model: Modèle FastText pré-entraîné.
        """
        super().__init__()
        self.type='text'
        self.model: FastText = model

    def compute(self, text1: str, text2: str) -> float:
        """
        Calcule la distance entre deux textes en utilisant les vecteurs de mots FastText.

        :param text1: Premier texte.
        :param text2: Deuxième texte.
        :return: Distance cosinus entre les deux textes.
        """
        # Obtenir la moyenne des vecteurs de mots pour chaque texte
        vector1: List[float] = self._get_sentence_vector(text1)
        vector2: List[float] = self._get_sentence_vector(text2)

        # Calcul de la similarité cosinus entre les deux vecteurs
        distance: float = Cosine().compute(vector1, vector2)
        
        return distance

    def _get_sentence_vector(self, text: str) -> List[float]:
        """
        Obtenir la moyenne des vecteurs FastText pour un texte donné.

        :param text: Texte à convertir en vecteur.
        :return: Vecteur moyen du texte.
        """
        words: List[str] = text.lower().split()
        word_vectors: List[List[float]] = [self.model.wv[word].tolist() for word in words if word in self.model.wv]

        # Si aucun mot n'est trouvé dans le modèle, retourner un vecteur nul de la taille appropriée
        if not word_vectors:
            return [0.0] * self.model.vector_size

        # Calculer la moyenne des vecteurs de mots
        sentence_vector: List[float] = [0.0] * self.model.vector_size
        for vector in word_vectors:
            for i in range(len(vector)):
                sentence_vector[i] += vector[i]

        # Diviser chaque élément par le nombre de vecteurs pour obtenir la moyenne
        sentence_vector = [v / len(word_vectors) for v in sentence_vector]
        return sentence_vector

class Dice(Distance):
	
	def __init__(self)-> None:
		super().__init__()
		self.type='vec_str'

	def compute(self,list1 :list, list2 :list)-> float:
		"""
		Calculate the Dice distance between two sets.
    
		:param set1: The first set (a set of elements)
		:param set2: The second set (a set of elements)
		:return: Dice distance between the two sets
		"""
		set1=set(list1)
		set2=set(list2)
		intersection = len(set1.intersection(set2))
		total_elements = len(set1) + len(set2)
    
		if total_elements == 0:
			return 0.0
    
		dice_coefficient = (2 * intersection) / total_elements
    
		return 1 - dice_coefficient
		
		
class Tversky(Distance):
	
	def __init__(self,alpha=0.5, beta=0.5)-> None:
		super().__init__()
		self.alpha=alpha
		self.beta=beta
		self.type='vec_str'
	def compute(self,set1 :set, set2 :set )-> float:
		"""
		Calcule la distance de Tversky entre deux ensembles.
    
		:param set1: Premier ensemble
		:param set2: Deuxième ensemble
		:param alpha: Paramètre de pondération pour |A - B|
		:param beta: Paramètre de pondération pour |B - A|
		:return: Distance de Tversky
		"""
		# Taille de l'intersection des ensembles
		intersection = len(set1 & set2)
    
		# Taille des éléments uniques à chaque ensemble
		unique_to_set1 = len(set1 - set2)
		unique_to_set2 = len(set2 - set1)
    
		# Calcul du coefficient de Tversky
		tversky_coeff = intersection / (intersection + self.alpha * unique_to_set1 + self.beta * unique_to_set2)
    
		# La distance est 1 moins le coefficient de Tversky
		distance = 1 - tversky_coeff
    
		return distance
		
	def example(self):
		self.obj1_example =  {'a', 'b', 'c', 'd'}
		self.obj2_example = {'c', 'd', 'e', 'f'}
		distance=self.compute(self.obj1_example,self.obj2_example)
		print(f"{self.__class__.__name__} distance between {self.obj1_example} and {self.obj2_example} is {distance:.2f}")
		
class NgramDistance(Distance):

    def __init__(self, n: int) -> None:
        """
        Initialise la classe avec la longueur des n-grams.

        :param n: Longueur des n-grams.
        """
        super().__init__()
        self.type='text'
        self.n: int = n

    def compute(self, text1: str, text2: str) -> float:
        """
        Calcule la distance N-gram entre deux textes.

        :param text1: Premier texte à comparer.
        :param text2: Deuxième texte à comparer.
        :return: Distance entre les deux textes basée sur les n-grams.
        """
        # Convertir les textes en ensembles de n-grams
        ngrams1: Set[str] = self._get_ngrams(text1)
        ngrams2: Set[str] = self._get_ngrams(text2)

        # Calcul de la distance Jaccard (intersection / union des n-grams)
        intersection_size: int = len(ngrams1 & ngrams2)
        union_size: int = len(ngrams1 | ngrams2)

        if union_size == 0:
            return 1.0  # Si l'union est vide, les textes sont complètement différents.

        # Distance de Jaccard (1 - similarité)
        distance: float = 1 - (intersection_size / union_size)
        return distance

    def _get_ngrams(self, text: str) -> Set[str]:
        """
        Génère les n-grams à partir d'un texte donné.

        :param text: Texte à convertir en n-grams.
        :return: Ensemble des n-grams du texte.
        """
        text = text.lower().replace(" ", "")  # Simplifier le texte
        ngrams: Set[str] = set()

        for i in range(len(text) - self.n + 1):
            ngram: str = text[i:i + self.n]
            ngrams.add(ngram)

        return ngrams

class SmithWaterman(Distance):

    def __init__(self, match_score: int, mismatch_penalty: int, gap_penalty: int) -> None:
        """
        Initialise la classe avec les paramètres de score de correspondance, de pénalité de mismatch et de pénalité de gap.

        :param match_score: Score attribué pour une correspondance entre deux caractères.
        :param mismatch_penalty: Pénalité pour une différence entre deux caractères.
        :param gap_penalty: Pénalité pour l'introduction d'un gap (trou).
        """
        super().__init__()
        self.type='text'
        self.match_score: int = match_score
        self.mismatch_penalty: int = mismatch_penalty
        self.gap_penalty: int = gap_penalty

    def compute(self, seq1: str, seq2: str) -> Tuple[int, List[List[int]]]:
        """
        Calcule l'alignement optimal entre deux chaînes à l'aide de l'algorithme Smith-Waterman.

        :param seq1: Première séquence à comparer.
        :param seq2: Deuxième séquence à comparer.
        :return: Le score d'alignement optimal et la matrice de score.
        """
        len1: int = len(seq1)
        len2: int = len(seq2)

        # Initialisation de la matrice de score avec des zéros
        score_matrix: List[List[int]] = [[0 for _ in range(len2 + 1)] for _ in range(len1 + 1)]

        # Variables pour suivre le score maximal
        max_score: int = 0

        # Remplissage de la matrice de score
        for i in range(1, len1 + 1):
            for j in range(1, len2 + 1):
                match: int = self.match_score if seq1[i - 1] == seq2[j - 1] else self.mismatch_penalty
                score_diag: int = score_matrix[i - 1][j - 1] + match
                score_up: int = score_matrix[i - 1][j] + self.gap_penalty
                score_left: int = score_matrix[i][j - 1] + self.gap_penalty
                score_matrix[i][j] = max(0, score_diag, score_up, score_left)

                # Mise à jour du score maximal
                max_score = max(max_score, score_matrix[i][j])

        return max_score, score_matrix

    def traceback(self, score_matrix: List[List[int]], seq1: str, seq2: str) -> Tuple[str, str]:
        """
        Effectue le traceback pour trouver l'alignement optimal.

        :param score_matrix: La matrice de score calculée par l'algorithme Smith-Waterman.
        :param seq1: Première séquence à aligner.
        :param seq2: Deuxième séquence à aligner.
        :return: Les deux séquences alignées.
        """
        align1: str = ""
        align2: str = ""

        # Trouver la position avec le score maximal
        max_i, max_j = 0, 0
        max_score: int = 0
        for i in range(len(score_matrix)):
            for j in range(len(score_matrix[0])):
                if score_matrix[i][j] > max_score:
                    max_score = score_matrix[i][j]
                    max_i, max_j = i, j

        i, j = max_i, max_j

        # Effectuer le traceback
        while i > 0 and j > 0 and score_matrix[i][j] > 0:
            if score_matrix[i][j] == score_matrix[i - 1][j - 1] + (self.match_score if seq1[i - 1] == seq2[j - 1] else self.mismatch_penalty):
                align1 = seq1[i - 1] + align1
                align2 = seq2[j - 1] + align2
                i -= 1
                j -= 1
            elif score_matrix[i][j] == score_matrix[i - 1][j] + self.gap_penalty:
                align1 = seq1[i - 1] + align1
                align2 = "-" + align2
                i -= 1
            else:
                align1 = "-" + align1
                align2 = seq2[j - 1] + align2
                j -= 1

        return align1, align2

class RatcliffObershelp(Distance):
	
	def __init__(self)-> None:
		super().__init__()
		self.type='text'
	def compute(self,str1 :str, str2 :str)-> float:
		"""
		ou Similarité Gestalt.
		Calculate the Ratcliff/Obershelp distance between two strings.
    
		:param str1: The first string
		:param str2: The second string
		:return: Ratcliff/Obershelp distance between the two strings
		"""
		def find_longest_common_substring(s1, s2):
			"""
			Helper function to find the longest common substring between two strings.
        
			:param s1: The first string
			:param s2: The second string
			:return: The longest common substring
			"""
			matrix = [[0] * (len(s2) + 1) for _ in range(len(s1) + 1)]
			longest_length = 0
			lcs_end = 0

			for i in range(1, len(s1) + 1):
				for j in range(1, len(s2) + 1):
					if s1[i - 1] == s2[j - 1]:
						matrix[i][j] = matrix[i - 1][j - 1] + 1
						if matrix[i][j] > longest_length:
							longest_length = matrix[i][j]
							lcs_end = i

			return s1[lcs_end - longest_length: lcs_end]

class Gestalt(RatcliffObershelp):
	def __init__(self)-> None:
		super().__init__()
		
from collections import Counter
import math

class BLEUScore(Distance):

    def __init__(self, n_gram: int = 4, smoothing: bool = True) -> None:
        """
        Initialise la classe BLEU Score avec le paramètre du n-gram et le choix de la méthode de lissage.

        :param n_gram: Le maximum n-gram à prendre en compte (par défaut 4).
        :param smoothing: Si le lissage est appliqué ou non (par défaut True).
        """
        super().__init__()
        self.type='text'
        self.n_gram = n_gram
        self.smoothing = smoothing

    def _extract_ngrams(self, text: List[str], n: int) -> List[Tuple[str]]:
        """
        Extrait les n-grams d'une liste de mots donnée.

        :param text: La liste des mots d'un texte.
        :param n: La longueur des n-grams à extraire.
        :return: Une liste de tuples représentant les n-grams.
        """
        return [tuple(text[i:i + n]) for i in range(len(text) - n + 1)]

    def _modified_precision(self, hypothesis: List[str], references: List[List[str]], n: int) -> float:
        """
        Calcule la précision modifiée pour un n-gram spécifique.

        :param hypothesis: La liste des mots dans le texte généré.
        :param references: La liste de listes de mots dans les textes de référence.
        :param n: La longueur du n-gram à prendre en compte.
        :return: La précision modifiée pour le n-gram.
        """
        hyp_ngrams = Counter(self._extract_ngrams(hypothesis, n))
        max_ref_ngrams = Counter()

        for reference in references:
            ref_ngrams = Counter(self._extract_ngrams(reference, n))
            for ngram in ref_ngrams:
                max_ref_ngrams[ngram] = max(max_ref_ngrams[ngram], ref_ngrams[ngram])

        clipped_count = 0
        total_count = sum(hyp_ngrams.values())

        for ngram in hyp_ngrams:
            clipped_count += min(hyp_ngrams[ngram], max_ref_ngrams[ngram])

        if total_count == 0:
            return 0.0

        return clipped_count / total_count

    def _brevity_penalty(self, hypothesis: List[str], references: List[List[str]]) -> float:
        """
        Calcule la pénalité de longueur (brevity penalty).

        :param hypothesis: La liste des mots dans le texte généré.
        :param references: La liste de listes de mots dans les textes de référence.
        :return: La pénalité de longueur.
        """
        hyp_len = len(hypothesis)
        ref_lens = [len(ref) for ref in references]
        closest_ref_len = min(ref_lens, key=lambda ref_len: (abs(ref_len - hyp_len), ref_len))

        if hyp_len > closest_ref_len:
            return 1.0
        elif hyp_len == 0:
            return 0.0
        else:
            return math.exp(1 - closest_ref_len / hyp_len)

    def compute(self, hypothesis: List[str], references: List[List[str]]) -> float:
        """
        Calcule le BLEU Score pour un texte généré donné par rapport à plusieurs textes de référence.

        :param hypothesis: La liste des mots dans le texte généré.
        :param references: La liste de listes de mots dans les textes de référence.
        :return: Le BLEU Score.
        """
        precisions: List[float] = []
        for n in range(1, self.n_gram + 1):
            precision = self._modified_precision(hypothesis, references, n)
            if self.smoothing and precision == 0:
                precision = 1e-9  # Smoothing pour éviter les zéros
            precisions.append(precision)

        # Moyenne géométrique des précisions
        score = math.exp(sum(math.log(p) for p in precisions) / self.n_gram)

        # Appliquer la pénalité de longueur (brevity penalty)
        bp = self._brevity_penalty(hypothesis, references)

        return score * bp

class ROUGEScore(Distance):

    def __init__(self, n_gram: int = 2) -> None:
        """
        Initializes the ROUGEScore class with a specified n-gram length.

        :param n_gram: The length of the n-gram to be considered (default is 2).
        """
        super().__init__()
        self.type='text'
        self.n_gram = n_gram

    def _extract_ngrams(self, text: List[str], n: int) -> List[Tuple[str]]:
        """
        Extracts n-grams from a given list of words.

        :param text: List of words from a text.
        :param n: The n-gram length to extract.
        :return: A list of n-grams (as tuples of words).
        """
        return [tuple(text[i:i + n]) for i in range(len(text) - n + 1)]

    def _ngram_overlap(self, reference: List[str], hypothesis: List[str], n: int) -> Tuple[int, int]:
        """
        Calculates the overlap of n-grams between the reference and the hypothesis text.

        :param reference: List of words from the reference text.
        :param hypothesis: List of words from the hypothesis text.
        :param n: The n-gram length to consider.
        :return: A tuple (overlap count, reference n-gram count).
        """
        ref_ngrams = Counter(self._extract_ngrams(reference, n))
        hyp_ngrams = Counter(self._extract_ngrams(hypothesis, n))

        overlap_count = 0
        for ngram in hyp_ngrams:
            overlap_count += min(hyp_ngrams[ngram], ref_ngrams.get(ngram, 0))

        return overlap_count, sum(ref_ngrams.values())

    def _recall(self, overlap: int, total: int) -> float:
        """
        Calculates the recall for a given overlap count.

        :param overlap: The number of overlapping n-grams between the reference and hypothesis.
        :param total: The total number of n-grams in the reference.
        :return: The recall score.
        """
        if total == 0:
            return 0.0
        return overlap / total

    def rouge_n(self, hypothesis: List[str], references: List[List[str]]) -> Dict[str, float]:
        """
        Computes the ROUGE-N score for the hypothesis against multiple references.

        :param hypothesis: List of words from the hypothesis text.
        :param references: List of lists of words from the reference texts.
        :return: A dictionary containing recall and F1 scores for n-grams.
        """
        recall_scores = []

        for reference in references:
            overlap, ref_count = self._ngram_overlap(reference, hypothesis, self.n_gram)
            recall = self._recall(overlap, ref_count)
            recall_scores.append(recall)

        best_recall = max(recall_scores)

        return {
            'recall': best_recall,
        }

class SoftCosineSimilarity(Distance):

    def __init__(self, term_similarity_matrix: Dict[Tuple[str, str], float]) -> None:
        """
        Initializes the SoftCosineSimilarity class with a term similarity matrix.

        :param term_similarity_matrix: A dictionary that stores pairwise term similarity scores.
        """
        super().__init__()
        self.type='text'
        self.term_similarity_matrix = term_similarity_matrix

    def _compute_term_frequency(self, document: List[str]) -> Dict[str, int]:
        """
        Computes the term frequency for each word in the document.

        :param document: A list of words from the document.
        :return: A dictionary with terms as keys and their frequencies as values.
        """
        term_freq: Dict[str, int] = {}
        for term in document:
            term_freq[term] = term_freq.get(term, 0) + 1
        return term_freq

    def _compute_magnitude(self, term_freq: Dict[str, int]) -> float:
        """
        Computes the magnitude of a document vector based on term frequencies.

        :param term_freq: A dictionary of term frequencies.
        :return: The magnitude of the document vector.
        """
        magnitude: float = 0.0
        for term, freq in term_freq.items():
            magnitude += freq ** 2
        return math.sqrt(magnitude)

    def soft_cosine(self, doc1: List[str], doc2: List[str]) -> float:
        """
        Computes the Soft Cosine Similarity between two documents.

        :param doc1: A list of words from the first document.
        :param doc2: A list of words from the second document.
        :return: The Soft Cosine Similarity score.
        """
        # Step 1: Compute term frequencies for both documents
        term_freq1 = self._compute_term_frequency(doc1)
        term_freq2 = self._compute_term_frequency(doc2)

        # Step 2: Calculate the numerator (dot product with similarity adjustment)
        numerator: float = 0.0
        for term1, freq1 in term_freq1.items():
            for term2, freq2 in term_freq2.items():
                similarity = self.term_similarity_matrix.get((term1, term2), 0.0)
                numerator += freq1 * freq2 * similarity

        # Step 3: Calculate magnitudes for both documents
        magnitude1: float = self._compute_magnitude(term_freq1)
        magnitude2: float = self._compute_magnitude(term_freq2)

        # Step 4: Calculate the denominator (product of magnitudes)
        denominator: float = magnitude1 * magnitude2

        # Step 5: Return the Soft Cosine Similarity
        if denominator == 0:
            return 0.0
        return numerator / denominator

from typing import List, Dict
from sklearn.decomposition import LatentDirichletAllocation as LDA
from sklearn.decomposition import TruncatedSVD as LSA
from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer

class TopicModelingDistance(Distance):

    def __init__(self, method: str = 'LDA', num_topics: int = 10) -> None:
        """
        Initializes the TopicModelingDistance class with the specified topic modeling method (LDA or LSA).

        :param method: The topic modeling method ('LDA' or 'LSA').
        :param num_topics: The number of topics to be generated by the model.
        """
        super().__init__()
        self.type='text'
        self.method: str = method
        self.num_topics: int = num_topics
        self.vectorizer = None
        self.model = None

    def fit(self, documents: List[str]) -> None:
        """
        Fits the topic model to the provided documents.

        :param documents: A list of text documents.
        """
        if self.method == 'LDA':
            self.vectorizer = CountVectorizer(stop_words='english')
            doc_term_matrix = self.vectorizer.fit_transform(documents)
            self.model = LDA(n_components=self.num_topics, random_state=42)
        elif self.method == 'LSA':
            self.vectorizer = TfidfVectorizer(stop_words='english')
            doc_term_matrix = self.vectorizer.fit_transform(documents)
            self.model = LSA(n_components=self.num_topics, random_state=42)
        
        # Fit the model to the document-term matrix
        self.model.fit(doc_term_matrix)

    def _get_topic_distribution(self, document: str) -> List[float]:
        """
        Computes the topic distribution for a single document.

        :param document: The document for which the topic distribution is computed.
        :return: A list representing the topic distribution.
        """
        doc_vector = self.vectorizer.transform([document])
        if self.method == 'LDA':
            return self.model.transform(doc_vector)[0].tolist()
        elif self.method == 'LSA':
            return self.model.transform(doc_vector)[0].tolist()

    def euclidean_distance(self, dist1: List[float], dist2: List[float]) -> float:
        """
        Computes the Euclidean distance between two topic distributions.

        :param dist1: The topic distribution of the first document.
        :param dist2: The topic distribution of the second document.
        :return: The Euclidean distance as a float.
        """
        distance: float = sum((a - b) ** 2 for a, b in zip(dist1, dist2)) ** 0.5
        return distance

    def topic_distance(self, doc1: str, doc2: str) -> float:
        """
        Computes the distance between two documents based on their topic distributions.

        :param doc1: The first document.
        :param doc2: The second document.
        :return: The distance between the two documents' topic distributions.
        """
        # Get topic distributions for both documents
        topic_dist1 = self._get_topic_distribution(doc1)
        topic_dist2 = self._get_topic_distribution(doc2)

        # Compute the distance (Euclidean distance)
        return self.euclidean_distance(topic_dist1, topic_dist2)


class AlignmentBasedMeasures(Distance):

    def __init__(self) -> None:
      super().__init__()
      self.type='text'
    def align_texts(self, text1: str, text2: str) -> List[Tuple[str, str]]:
        """
        Aligns two texts at the word level by using dynamic programming to find the optimal alignment.

        :param text1: The first text as a string.
        :param text2: The second text as a string.
        :return: A list of tuples representing the aligned words from both texts.
        """
        words1: List[str] = text1.split()
        words2: List[str] = text2.split()

        # Initialize DP table for alignment
        dp: List[List[int]] = [[0] * (len(words2) + 1) for _ in range(len(words1) + 1)]
        alignment: List[Tuple[str, str]] = []

        # Fill DP table (Levenshtein-like)
        for i in range(1, len(words1) + 1):
            dp[i][0] = i
        for j in range(1, len(words2) + 1):
            dp[0][j] = j

        for i in range(1, len(words1) + 1):
            for j in range(1, len(words2) + 1):
                cost = 0 if words1[i - 1] == words2[j - 1] else 1
                dp[i][j] = min(dp[i - 1][j] + 1,    # Deletion
                               dp[i][j - 1] + 1,    # Insertion
                               dp[i - 1][j - 1] + cost)  # Substitution

        # Backtrack to find the alignment
        i, j = len(words1), len(words2)
        while i > 0 or j > 0:
            if i > 0 and j > 0 and dp[i][j] == dp[i - 1][j - 1] + (0 if words1[i - 1] == words2[j - 1] else 1):
                alignment.append((words1[i - 1], words2[j - 1]))
                i -= 1
                j -= 1
            elif i > 0 and dp[i][j] == dp[i - 1][j] + 1:
                alignment.append((words1[i - 1], "-"))  # Deletion
                i -= 1
            else:
                alignment.append(("-", words2[j - 1]))  # Insertion
                j -= 1

        alignment.reverse()  # Reverse to get the correct order
        return alignment

    def alignment_score(self, text1: str, text2: str) -> float:
        """
        Computes an alignment score based on the number of aligned words.

        :param text1: The first text.
        :param text2: The second text.
        :return: A float representing the alignment score (higher is better).
        """
        aligned: List[Tuple[str, str]] = self.align_texts(text1, text2)
        matches: int = sum(1 for word1, word2 in aligned if word1 == word2)
        total_words: int = max(len(text1.split()), len(text2.split()))
        
        return matches / total_words


class GappyNGramDistance(Distance):

    def __init__(self, n: int, gap_size: int) -> None:
        """
        Initializes the GappyNGramDistance class with the given n-gram length and gap size.

        :param n: The length of the n-gram.
        :param gap_size: The size of the allowed gap between the elements of the n-gram.
        """
        super().__init__()
        self.type='text'
        self.n: int = n
        self.gap_size: int = gap_size

    def generate_gappy_ngrams(self, text: str) -> Set[Tuple[str, ...]]:
        """
        Generates the gappy n-grams for a given text.

        :param text: The input text.
        :return: A set of gappy n-grams, where each n-gram is represented as a tuple of strings.
        """
        words: List[str] = text.split()
        gappy_ngrams: Set[Tuple[str, ...]] = set()

        for i in range(len(words) - self.n + 1):
            for j in range(1, self.gap_size + 2):
                if i + j * (self.n - 1) < len(words):
                    gappy_ngram: Tuple[str, ...] = tuple(words[i + k * j] for k in range(self.n))
                    gappy_ngrams.add(gappy_ngram)

        return gappy_ngrams

    def compute(self, text1: str, text2: str) -> float:
        """
        Computes the Gappy N-gram similarity between two texts.

        :param text1: The first text.
        :param text2: The second text.
        :return: The similarity score between the two sets of gappy n-grams (from 0 to 1).
        """
        ngrams1: Set[Tuple[str, ...]] = self.generate_gappy_ngrams(text1)
        ngrams2: Set[Tuple[str, ...]] = self.generate_gappy_ngrams(text2)

        intersection: Set[Tuple[str, ...]] = ngrams1.intersection(ngrams2)
        union: Set[Tuple[str, ...]] = ngrams1.union(ngrams2)

        if not union:  # Handle division by zero when both n-gram sets are empty
            return 1.0 if not intersection else 0.0

        return len(intersection) / len(union)

class SoftJaccardSimilarity(Distance):

    def __init__(self, threshold: float) -> None:
        """
        Initializes the Soft Jaccard Similarity class with a similarity threshold.

        :param threshold: The minimum similarity between two words to be considered as a match (between 0 and 1).
        """
        super().__init__()
        self.type='text'
        self.threshold: float = threshold

    def word_similarity(self, word1: str, word2: str) -> float:
        """
        Calculates the similarity between two words based on their character-level similarity.
        This function can be replaced with more complex similarity measures (e.g., Levenshtein distance, cosine similarity).

        :param word1: The first word.
        :param word2: The second word.
        :return: A similarity score between 0 and 1.
        """
        set1: Set[str] = set(word1)
        set2: Set[str] = set(word2)

        intersection_size: int = len(set1.intersection(set2))
        union_size: int = len(set1.union(set2))

        return intersection_size / union_size if union_size != 0 else 0.0

    def compute(self, text1: str, text2: str) -> float:
        """
        Calculates the Soft Jaccard Similarity between two texts.

        :param text1: The first text.
        :param text2: The second text.
        :return: The soft Jaccard similarity score between the two sets of words (between 0 and 1).
        """
        words1: List[str] = text1.split()
        words2: List[str] = text2.split()

        matched_words: Set[Tuple[str, str]] = set()
        for word1 in words1:
            for word2 in words2:
                if self.word_similarity(word1, word2) >= self.threshold:
                    matched_words.add((word1, word2))

        # Intersection is the number of matched pairs
        intersection_size: int = len(matched_words)
        # Union is the total number of unique words from both sets
        union_size: int = len(set(words1).union(set(words2)))

        return intersection_size / union_size if union_size != 0 else 0.0

import zlib

class NormalizedCompressionDistance(Distance):

    def __init__(self) -> None:
      super().__init__()
      self.type='text'
    def compress(self, text: str) -> int:
        """
        Compresses the input text using zlib and returns the length of the compressed data.

        :param text: The input text to be compressed.
        :return: The length of the compressed text in bytes.
        """
        compressed_data = zlib.compress(text.encode('utf-8'))
        return len(compressed_data)

    def ncd(self, text1: str, text2: str) -> float:
        """
        Calculates the Normalized Compression Distance (NCD) between two texts.

        :param text1: The first text.
        :param text2: The second text.
        :return: The NCD value between 0 and 1, where lower values indicate more similarity.
        """
        C_x = self.compress(text1)
        C_y = self.compress(text2)
        C_xy = self.compress(text1 + text2)

        return (C_xy - min(C_x, C_y)) / max(C_x, C_y)

class MongeElkanDistance(Distance):
    def __init__(self, base_distance: Callable[[str, str], float]) -> None:
        """
        Initialise la classe avec une fonction de distance de base.

        :param base_distance: Fonction de distance de base entre deux sous-éléments (par ex. Levenshtein).
        """
        super().__init__()
        self.type='text'
        self.base_distance: Callable[[str, str], float] = base_distance

    def compute(self, set1: List[str], set2: List[str]) -> float:
        """
        Calcule la distance Monge-Elkan entre deux ensembles de mots ou sous-éléments.

        :param set1: Premier ensemble de mots (texte décomposé en liste de mots ou caractères).
        :param set2: Deuxième ensemble de mots.
        :return: La distance Monge-Elkan calculée.
        """
        if not set1 or not set2:
            return float('inf')

        total_distance: float = 0.0
        for word1 in set1:
            min_distance: float = min(self.base_distance(word1, word2) for word2 in set2)
            total_distance += min_distance

        return total_distance / len(set1)

from typing import List, Dict
import math
from collections import Counter

class JensenShannonDivergence(Distance):
    def __init__(self) -> None:
        """
        Initialise la classe avec une fonction de distance de base.

        :param base_distance: Fonction de distance de base entre deux sous-éléments (par ex. Levenshtein).
        """
        super().__init__()
        self.type='text'
    def compute(self, dist1: List[float], dist2: List[float]) -> float:
        """
        Calcule la Jensen-Shannon Divergence entre deux distributions de probabilités.

        :param dist1: Première distribution de probabilités (somme égale à 1).
        :param dist2: Deuxième distribution de probabilités (somme égale à 1).
        :return: La divergence Jensen-Shannon entre les deux distributions.
        """
        if len(dist1) != len(dist2):
            raise ValueError("Les distributions doivent avoir la même longueur")

        # Calcul de la distribution moyenne
        avg_dist: List[float] = [(p1 + p2) / 2 for p1, p2 in zip(dist1, dist2)]

        # Calcul de la divergence KL pour les deux distributions par rapport à la distribution moyenne
        kl_div1: float = self._kl_divergence(dist1, avg_dist)
        kl_div2: float = self._kl_divergence(dist2, avg_dist)

        # La Jensen-Shannon Divergence est la moyenne des deux divergences KL
        return (kl_div1 + kl_div2) / 2

    def _kl_divergence(self, dist_p: List[float], dist_q: List[float]) -> float:
        """
        Calcule la Kullback-Leibler Divergence entre deux distributions.

        :param dist_p: Distribution de probabilité p.
        :param dist_q: Distribution de probabilité q.
        :return: La divergence KL entre les distributions p et q.
        """
        divergence: float = 0.0
        for p, q in zip(dist_p, dist_q):
            if p > 0 and q > 0:
                divergence += p * math.log(p / q)
        return divergence

    def text_to_distribution(self, text: str, vocabulary: List[str]) -> List[float]:
        """
        Convertit un texte en distribution de probabilités basée sur la fréquence des mots dans un vocabulaire.

        :param text: Le texte à convertir.
        :param vocabulary: La liste des mots qui composent le vocabulaire.
        :return: Une distribution de probabilités (liste de fréquences normalisées).
        """
        # Tokenize le texte
        word_counts: Dict[str, int] = Counter(text.split())

        # Crée une distribution basée sur le vocabulaire
        dist: List[float] = [word_counts.get(word, 0) for word in vocabulary]

        # Normalise la distribution pour que la somme soit égale à 1
        total_count: float = sum(dist)
        if total_count > 0:
            dist = [count / total_count for count in dist]

        return dist





cdef class Distance:
	pass

cdef class Euclidean:

	cpdef float compute(self,list[float] point1, list[float] point2) 

	cpdef void exemple(self)
    
cdef class Levenshtein:

	cpdef int compute(self, str s1, str s2)except *

	cpdef void exemple(self)

cdef class Hamming:

	cpdef int compute(self, str str1, str str2)

	cpdef void exemple(self)

cdef class InverseTanimoto:
	
	cpdef double compute(self, set set_a, set set_b)
	
	cpdef void exemple(self)

cdef class Manhattan:
	
	cpdef double compute(self, list[float] point1, list[float] point2)
	
	cpdef void exemple(self)

cdef class Minkowski:
	
	cpdef double compute(self, list[float] point1, list[float] point2)
	
	cpdef void exemple(self)

cdef class Mahalanobis:
	
	cpdef double mean(self, list[list[float]] data)
	
	cpdef double covariance_matrix(self, list[list[float]] data,float mean)
	
	cpdef list[list[float]] matrix_inverse(self, list[list[float]] matrix)
	
	cpdef double compute(self, list[float] point, list[list[float]] data)
	cpdef void exemple(self)

cdef class RussellRao:
	
	cpdef double compute(self, list[int] vector1, list[int]vector2)
	
	cpdef void exemple(self)

cdef class Chebyshev:
	
	cpdef double compute(self, list[float] point1, list[loat]point2)
	
	cpdef void exemple(self)

cdef class Jaro:
	
	cpdef double compute(self, str s1, str s2)
	
	cpdef void exemple(self)

cdef class SorensenDice:
	
	cpdef double compute(self, str s1, str s2)
	
	cpdef void exemple(self)

cdef class JaroWinkler:
	
	cpdef double compute(self, str s1, str s2,float p)
	
	cpdef void exemple(self)

cdef class Hausdorff:
	
	cpdef double compute(self, set set_a, set set_b)
	
	cpdef void exemple(self)

cdef class KendallTau:
	
	cpdef double compute(self, list[int] permutation1, list[int] permutation1)
	
	cpdef void exemple(self)

cdef class Haversine:
	
	cpdef double compute(self, list[float] p1, list[loat]p2)
	
	cpdef void exemple(self)

cdef class Canberra:
	
	cpdef double compute(self, list[float] point1, list[loat]point2)
	
	cpdef void exemple(self)

cdef class BrayCurtis:
	
	cpdef double compute(self, list[float] point1, list[loat]point2)
	
	cpdef void exemple(self)

cdef class Matching:
	
	cpdef int compute(self, str seq1, str seq2)
	
	cpdef void exemple(self)

cdef class Kulsinski:
	
	cpdef double compute(self, set set_a, set set_b)
	
	cpdef void exemple(self)

cdef class DamerauLevenshtein:
	
	cpdef int compute(self, str s1, str s2)
	
	cpdef void exemple(self)

cdef class Yule:
	
	cpdef double compute(self, list[int] binary_vector1, list[int]binary_vector2)
	
	cpdef void exemple(self)

cdef class Bhattacharyya:
	
	cpdef double compute(self, list[float] P, list[float]Q)
	
	cpdef void exemple(self)

cdef class Gower:
	
	cpdef double compute(self, list vec1, list vec2)
	
	cpdef void exemple(self)

cdef class Hellinger:
	
	cpdef double compute(self, list[float] p, list[float]q)
	
	cpdef void exemple(self)

cdef class CzekanowskiDice:
	
	cpdef double compute(self, list[float] x, list[float]y)
	
	cpdef void exemple(self)

cdef class Wasserstein:
	
	cpdef double compute(self, list[float] distribution1, list[float]distribution2)
	
	cpdef array double _cumulative_distribution(list[float] distribution)
	
	cpdef void exemple(self)







import setuptools
#from setuptools import setup, Extension,find_packages

from distutils.core import setup, Extension

from Cython.Build import cythonize
from readme_renderer import markdown
#ext_modules = cythonize([Extension("distancia.distance", ["distancia/distance.py"])])

with open("README.md", "r") as fh:
    long_description = fh.read()
#with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
#    long_description = f.read()    
 

setuptools.setup(
#setup(
    
    #ext_modules=ext_modules,
    #ext_modules = cythonize(ext_modules),


    name="distancia", # Replace with your username

    version="0.0.63",

    author="Yves Mercadier",

    author_email="",

    description="distance metrics,data-science deep-learning machine-learning neural-network",

    long_description=markdown.render(long_description),
    #long_description=long_description,

    long_description_content_type="text/markdown",

    url="https://pypi.org/project/distancia/",

    packages=setuptools.find_packages(),

    classifiers=[

        "Programming Language :: Python :: 3",

        "License :: OSI Approved :: MIT License",

        "Operating System :: OS Independent",

    ],

    python_requires='>=3.0',

)



from .mainClass import *


import cmath
from typing import List, Tuple

class SpectralConvergence(Distance):

    def __init__(self) -> None:
        super().__init__()

    def fft(self, signal: List[float]) -> List[complex]:
        """
        Compute the Fast Fourier Transform (FFT) of the input signal.

        :param signal: The input signal as a list of floats.
        :return: The FFT of the signal as a list of complex numbers.
        """
        n: int = len(signal)
        if n == 1:
            return signal
        else:
            even: List[complex] = self.fft(signal[0::2])
            odd: List[complex] = self.fft(signal[1::2])
            combined: List[complex] = [0] * n
            for k in range(n // 2):
                t: complex = cmath.exp(-2j * cmath.pi * k / n) * odd[k]
                combined[k] = even[k] + t
                combined[k + n // 2] = even[k] - t
            return combined

    def magnitude(self, spectrum: List[complex]) -> List[float]:
        """
        Compute the magnitude of a complex spectrum.

        :param spectrum: A list of complex numbers representing the frequency spectrum.
        :return: A list of floats representing the magnitude of the spectrum.
        """
        return [abs(value) for value in spectrum]

    def compute(self, signal1: List[float], signal2: List[float]) -> float:
        """
        Compute the Spectral Convergence between two signals.

        :param signal1: The first signal as a list of floats.
        :param signal2: The second signal as a list of floats.
        :return: The spectral convergence between the two signals as a float.
        """
        # Compute the FFT for both signals
        fft1: List[complex] = self.fft(signal1)
        fft2: List[complex] = self.fft(signal2)

        # Compute magnitudes of the spectrums
        mag1: List[float] = self.magnitude(fft1)
        mag2: List[float] = self.magnitude(fft2)

        # Ensure both spectrums are of the same length
        if len(mag1) != len(mag2):
            raise ValueError("Both signals must have the same length.")

        # Compute the Spectral Convergence
        numerator: float = sum(abs(m1 - m2) for m1, m2 in zip(mag1, mag2))
        denominator: float = sum(mag1)

        # To avoid division by zero
        if denominator == 0:
            return float('inf')

        return numerator / denominator

import math
import cmath

class MFCCProcessor(Distance):
    def __init__(self, sample_rate: int = 16000, n_mfcc: int = 13, n_fft: int = 2048, n_mels: int = 26)-> None:
        super().__init__()

        self.sample_rate = sample_rate
        self.n_mfcc = n_mfcc
        self.n_fft = n_fft
        self.n_mels = n_mels

    def _fft(self, signal: List[float]) -> List[complex]:
        n = len(signal)
        if n <= 1:
            return signal
        even = self._fft(signal[0::2])
        odd = self._fft(signal[1::2])
        combined = [0] * n
        for k in range(n // 2):
            t = cmath.exp(-2j * math.pi * k / n) * odd[k]
            combined[k] = even[k] + t
            combined[k + n // 2] = even[k] - t
        return combined

    def _mel_to_hz(self, mel: float) -> float:
        return 700 * (10 ** (mel / 2595) - 1)

    def _hz_to_mel(self, hz: float) -> float:
        return 2595 * math.log10(1 + hz / 700)

    def _mel_filterbank(self) -> List[List[float]]:
        low_freq_mel = self._hz_to_mel(0)
        high_freq_mel = self._hz_to_mel(self.sample_rate / 2)
        mel_points = [low_freq_mel + i * (high_freq_mel - low_freq_mel) / (self.n_mels + 1) for i in range(self.n_mels + 2)]
        hz_points = [self._mel_to_hz(mel) for mel in mel_points]
        bin_points = [int(round((self.n_fft + 1) * hz / self.sample_rate)) for hz in hz_points]

        fbank = [[0.0] * (self.n_fft // 2 + 1) for _ in range(self.n_mels)]
        for m in range(1, self.n_mels + 1):
            f_m_minus = bin_points[m - 1]
            f_m = bin_points[m]
            f_m_plus = bin_points[m + 1]

            for k in range(f_m_minus, f_m):
                fbank[m-1][k] = (k - f_m_minus) / (f_m - f_m_minus)
            for k in range(f_m, f_m_plus):
                fbank[m-1][k] = (f_m_plus - k) / (f_m_plus - f_m)

        return fbank

    def _dct(self, x: List[float]) -> List[float]:
        N = len(x)
        y = [0.0] * N
        for k in range(N):
            for n in range(N):
                y[k] += x[n] * math.cos(math.pi * k * (2 * n + 1) / (2 * N))
        return y

    def compute_mfcc(self, signal1: List[float], signal2: List[float]) -> Tuple[List[List[float]], List[List[float]]]:
        """
        Calcule les MFCC pour deux signaux audio.

        Args:
            signal1 (List[float]): Premier signal audio.
            signal2 (List[float]): Deuxième signal audio.

        Returns:
            Tuple[List[List[float]], List[List[float]]]: MFCC des deux signaux.
        """
        def process_signal(signal: List[float]) -> List[List[float]]:
            # Pré-accentuation
            pre_emphasis = 0.97
            emphasized_signal = [signal[i] - pre_emphasis * signal[i-1] for i in range(1, len(signal))]

            # Fenêtrage
            frame_length = self.n_fft
            frame_step = frame_length // 2
            frames = [emphasized_signal[i:i+frame_length] for i in range(0, len(emphasized_signal) - frame_length + 1, frame_step)]

            # Appliquer la fenêtre de Hamming
            hamming = [0.54 - 0.46 * math.cos(2 * math.pi * i / (frame_length - 1)) for i in range(frame_length)]
            windowed_frames = [[frame[i] * hamming[i] for i in range(len(frame))] for frame in frames]

            # FFT
            magnitude_frames = [[abs(x) for x in self._fft(frame)] for frame in windowed_frames]

            # Mel filterbank
            mel_fb = self._mel_filterbank()
            mel_spectrum = [[sum(m * f for m, f in zip(mel_filter, frame[:len(mel_filter)])) for mel_filter in mel_fb] for frame in magnitude_frames]

            # Log
            log_mel_spectrum = [[math.log(x + 1e-8) for x in frame] for frame in mel_spectrum]

            # DCT
            mfcc = [self._dct(frame)[:self.n_mfcc] for frame in log_mel_spectrum]

            return mfcc

        mfcc1 = process_signal(signal1)
        mfcc2 = process_signal(signal2)

        return mfcc1, mfcc2

    def compare_mfcc(self, signal1: List[float], signal2: List[float]) -> List[float]:
        """
        Calcule et compare les MFCC de deux signaux audio.

        Args:
            signal1 (List[float]): Premier signal audio.
            signal2 (List[float]): Deuxième signal audio.

        Returns:
            List[float]: Distance euclidienne moyenne entre les MFCC des deux signaux.
        """
        mfcc1, mfcc2 = self.compute_mfcc(signal1, signal2)

        # Assurez-vous que les deux MFCC ont le même nombre de trames
        min_frames = min(len(mfcc1), len(mfcc2))
        mfcc1 = mfcc1[:min_frames]
        mfcc2 = mfcc2[:min_frames]

        # Calculez la distance euclidienne moyenne
        distances = []
        for frame1, frame2 in zip(mfcc1, mfcc2):
            distance = math.sqrt(sum((a - b) ** 2 for a, b in zip(frame1, frame2)))
            distances.append(distance)

        return sum(distances) / len(distances)
        
#claude ai fft
from typing import List
import cmath

class SignalProcessor(Distance):

    def __init__(self) -> None:
        super().__init__()
    def _fft(self, signal: List[float]) -> List[complex]:
        """
        Calcule la Transformée de Fourier Rapide (FFT) d'un signal sonore.

        Args:
            signal (List[float]): Le signal d'entrée sous forme de liste de nombres flottants.

        Returns:
            List[complex]: La FFT du signal sous forme de liste de nombres complexes.
        """
        n = len(signal)
        if n <= 1:
            return signal

        # Diviser le signal en pair et impair
        even = self._fft(signal[0::2])
        odd = self._fft(signal[1::2])

        # Combiner
        combined = [0] * n
        for k in range(n // 2):
            t = cmath.exp(-2j * cmath.pi * k / n) * odd[k]
            combined[k] = even[k] + t
            combined[k + n // 2] = even[k] - t

        return combined

    @staticmethod
    def pad_to_power_of_two(signal: List[float]) -> List[float]:
        """
        Complète le signal avec des zéros pour atteindre une longueur qui est une puissance de 2.

        Args:
            signal (List[float]): Le signal d'entrée.

        Returns:
            List[float]: Le signal complété.
        """
        n = 1
        while n < len(signal):
            n *= 2
        return signal + [0.0] * (n - len(signal))
'''
processor = SignalProcessor()
signal1 = [0.1, 0.2, 0.3, 0.4, 0.5]  # exemple de signal
signal2 = [0.2, 0.3, 0.4, 0.5, 0.6]  # autre exemple de signal
fft_difference = processor._fft(signal1)
print(fft_difference)
'''
##############"
import math
from typing import List

class PowerSpectralDensityDistance(Distance):

    def __init__(self, sample_rate: int) -> None:
        super().__init__()

        self.sample_rate = sample_rate

    def _psd(self, signal: List[float]) -> List[float]:
        fft_result = SignalProcessor()._fft(signal)
        magnitude_spectrum = [abs(freq) ** 2 for freq in fft_result[:len(fft_result) // 2]]
        return magnitude_spectrum

    def compute_psd_distance(self, signal1: List[float], signal2: List[float]) -> float:
        psd1 = self._psd(signal1)
        psd2 = self._psd(signal2)

        distance = sum((psd1[i] - psd2[i]) ** 2 for i in range(min(len(psd1), len(psd2))))
        return math.sqrt(distance)
        
import math
from typing import List

class CrossCorrelation(Distance):

    def __init__(self, sample_rate: int) -> None:
        super().__init__()

        self.sample_rate: int = sample_rate

    def _mean(self, signal: List[float]) -> float:
        return sum(signal) / len(signal)

    def _normalize(self, signal: List[float]) -> List[float]:
        mean_value: float = self._mean(signal)
        return [x - mean_value for x in signal]

    def compute_cross_correlation(self, signal1: List[float], signal2: List[float]) -> float:
        signal1_normalized: List[float] = self._normalize(signal1)
        signal2_normalized: List[float] = self._normalize(signal2)

        numerator: float = sum(signal1_normalized[i] * signal2_normalized[i] for i in range(min(len(signal1_normalized), len(signal2_normalized))))
        denominator_signal1: float = math.sqrt(sum(x ** 2 for x in signal1_normalized))
        denominator_signal2: float = math.sqrt(sum(x ** 2 for x in signal2_normalized))

        denominator: float = denominator_signal1 * denominator_signal2

        return numerator / denominator if denominator != 0 else 0.0
        
#ai claude

from typing import List, Tuple
import math
import cmath

class PhaseDifferenceCalculator(Distance):

    def __init__(self, sample_rate: int, window_size: int, hop_size: int) -> None:
        """
        Initialise le calculateur de différence de phase.

        Args:
            sample_rate (int): Taux d'échantillonnage des signaux.
            window_size (int): Taille de la fenêtre pour l'analyse.
            hop_size (int): Taille du saut entre les fenêtres successives.
        """
        super().__init__()

        self.sample_rate: int = sample_rate
        self.window_size: int = window_size
        self.hop_size: int = hop_size

    def _apply_window(self, segment: List[float]) -> List[float]:
        """
        Applique une fenêtre de Hann au segment.

        Args:
            segment (List[float]): Segment du signal.

        Returns:
            List[float]: Segment après application de la fenêtre.
        """
        return [s * 0.5 * (1 - math.cos(2 * math.pi * i / (len(segment) - 1)))
                for i, s in enumerate(segment)]

    def _fft(self, signal: List[float]) -> List[complex]:
        """
        Calcule la Transformée de Fourier Rapide (FFT) du signal.

        Args:
            signal (List[float]): Signal d'entrée.

        Returns:
            List[complex]: FFT du signal.
        """
        n: int = len(signal)
        if n <= 1:
            return signal
        even: List[complex] = self._fft(signal[0::2])
        odd: List[complex] = self._fft(signal[1::2])
        combined: List[complex] = [0] * n
        for k in range(n // 2):
            t: complex = cmath.exp(-2j * math.pi * k / n) * odd[k]
            combined[k] = even[k] + t
            combined[k + n // 2] = even[k] - t
        return combined

    def calculate_phase_difference(self, signal1: List[float], signal2: List[float]) -> List[float]:
        """
        Calcule la différence de phase entre deux signaux.

        Args:
            signal1 (List[float]): Premier signal.
            signal2 (List[float]): Deuxième signal.

        Returns:
            List[float]: Différence de phase pour chaque segment.
        """
        if len(signal1) != len(signal2):
            raise ValueError("Les signaux doivent avoir la même longueur")

        phase_differences: List[float] = []
        num_segments: int = (len(signal1) - self.window_size) // self.hop_size + 1

        for i in range(num_segments):
            start: int = i * self.hop_size
            end: int = start + self.window_size

            segment1: List[float] = self._apply_window(signal1[start:end])
            segment2: List[float] = self._apply_window(signal2[start:end])

            fft1: List[complex] = self._fft(segment1)
            fft2: List[complex] = self._fft(segment2)

            phase_diff: float = 0
            for f1, f2 in zip(fft1, fft2):
                if abs(f1) > 1e-6 and abs(f2) > 1e-6:  # Éviter la division par zéro
                    phase1: float = cmath.phase(f1)
                    phase2: float = cmath.phase(f2)
                    diff: float = phase2 - phase1
                    # Normaliser la différence de phase entre -pi et pi
                    phase_diff += (diff + math.pi) % (2 * math.pi) - math.pi

            phase_differences.append(phase_diff / len(fft1))

        return phase_differences

    def get_time_axis(self) -> List[float]:
        """
        Génère l'axe temporel pour les différences de phase calculées.

        Returns:
            List[float]: Axe temporel en secondes.
        """
        num_segments: int = len(self.calculate_phase_difference([0] * self.window_size, [0] * self.window_size))
        return [i * self.hop_size / self.sample_rate for i in range(num_segments)]

    def analyze_signals(self, signal1: List[float], signal2: List[float]) -> Tuple[List[float], List[float]]:
        """
        Analyse deux signaux et retourne la différence de phase et l'axe temporel.

        Args:
            signal1 (List[float]): Premier signal.
            signal2 (List[float]): Deuxième signal.

        Returns:
            Tuple[List[float], List[float]]: Différence de phase et axe temporel.
        """
        phase_diff: List[float] = self.calculate_phase_difference(signal1, signal2)
        time_axis: List[float] = self.get_time_axis()
        return phase_diff, time_axis
        
from typing import List
import math

class TimeLagDistance(Distance):

    def __init__(self, sample_rate: int) -> None:
        super().__init__()

        self.sample_rate: int = sample_rate

    def _cross_correlation(self, signal1: List[float], signal2: List[float], lag: int) -> float:
        n: int = len(signal1)
        if lag > 0:
            shifted_signal2: List[float] = [0] * lag + signal2[:-lag]
        else:
            shifted_signal2: List[float] = signal2[-lag:] + [0] * (-lag)

        return sum(signal1[i] * shifted_signal2[i] for i in range(n))

    def compute_time_lag_distance(self, signal1: List[float], signal2: List[float], max_lag: int) -> int:
        best_lag: int = 0
        best_correlation: float = -float('inf')

        for lag in range(-max_lag, max_lag + 1):
            correlation: float = self._cross_correlation(signal1, signal2, lag)
            if correlation > best_correlation:
                best_correlation = correlation
                best_lag = lag

        return best_lag
        
from typing import List

class PESQ(Distance):

    def __init__(self, sample_rate: int) -> None:
        super().__init__()

        self.sample_rate: int = sample_rate

    def _preprocess(self, signal: List[float]) -> List[float]:
        # Placeholder preprocessing steps: normalization and filtering
        max_val: float = max(abs(x) for x in signal)
        return [x / max_val for x in signal] if max_val != 0 else signal

    def _compare_signals(self, reference: List[float], degraded: List[float]) -> float:
        # Placeholder function to simulate signal comparison
        mse: float = sum((reference[i] - degraded[i]) ** 2 for i in range(min(len(reference), len(degraded))))
        return mse / len(reference)

    def compute_pesq(self, reference_signal: List[float], degraded_signal: List[float]) -> float:
        reference_processed: List[float] = self._preprocess(reference_signal)
        degraded_processed: List[float] = self._preprocess(degraded_signal)

        comparison_score: float = self._compare_signals(reference_processed, degraded_processed)

        # Placeholder formula for PESQ score (the actual PESQ model is more complex)
        pesq_score: float = 4.5 - comparison_score  # 4.5 is the best score in PESQ scale

        return max(1.0, min(pesq_score, 4.5))  # PESQ scores typically range between 1.0 and 4.5
        
        
import cmath
from typing import List

def fft( signal: List[float]) -> List[complex]:
        """
        Compute the Fast Fourier Transform (FFT) of the input signal.

        :param signal: The input signal as a list of floats.
        :return: The FFT of the signal as a list of complex numbers.
        """
        n: int = len(signal)
        if n == 1:
            return signal
        else:
            even: List[complex] = fft(signal[0::2])
            odd: List[complex] = fft(signal[1::2])
            combined: List[complex] = [0] * n
            for k in range(n // 2):
                t: complex = cmath.exp(-2j * cmath.pi * k / n) * odd[k]
                combined[k] = even[k] + t
                combined[k + n // 2] = even[k] - t
            return combined

class LogSpectralDistance(Distance):

    def __init__(self, sample_rate: int) -> None:
        super().__init__()

        self.sample_rate: int = sample_rate

    def _log_magnitude_spectrum(self, signal: List[float]) -> List[float]:
        fft_result: List[complex] = fft(signal)
        return [20 * math.log10(abs(x)) if abs(x) != 0 else 0 for x in fft_result]

    def compute_lsd(self, signal1: List[float], signal2: List[float]) -> float:
        log_spectrum1: List[float] = self._log_magnitude_spectrum(signal1)
        log_spectrum2: List[float] = self._log_magnitude_spectrum(signal2)

        # Calculate the squared differences between the log-magnitude spectra
        squared_diffs: List[float] = [(log_spectrum1[i] - log_spectrum2[i]) ** 2 for i in range(min(len(log_spectrum1), len(log_spectrum2)))]

        # Compute the LSD value
        mean_squared_diff: float = sum(squared_diffs) / len(squared_diffs)
        return math.sqrt(mean_squared_diff)
        
import math
from typing import List

def fft( signal: List[float]) -> List[complex]:
        """
        Compute the Fast Fourier Transform (FFT) of the input signal.

        :param signal: The input signal as a list of floats.
        :return: The FFT of the signal as a list of complex numbers.
        """
        n: int = len(signal)
        if n == 1:
            return signal
        else:
            even: List[complex] = fft(signal[0::2])
            odd: List[complex] = fft(signal[1::2])
            combined: List[complex] = [0] * n
            for k in range(n // 2):
                t: complex = cmath.exp(-2j * cmath.pi * k / n) * odd[k]
                combined[k] = even[k] + t
                combined[k + n // 2] = even[k] - t
            return combined

class BarkSpectralDistortion(Distance):

    def __init__(self, sample_rate: int) -> None:
        super().__init__()

        self.sample_rate: int = sample_rate

    def _bark_scale(self, freq: float) -> float:
        return 13 * math.atan(0.00076 * freq) + 3.5 * math.atan((freq / 7500) ** 2)

    def _compute_bark_spectrum(self, signal: List[float]) -> List[float]:
        fft_result: List[complex] = fft(signal)
        N: int = len(fft_result)
        bark_spectrum: List[float] = [0.0] * N

        for i in range(N):
            freq: float = i * (self.sample_rate / N)
            bark_freq: float = self._bark_scale(freq)
            magnitude: float = abs(fft_result[i])
            bark_spectrum[i] = 20 * math.log10(magnitude) if magnitude != 0 else 0

        return bark_spectrum

    def compute_bsd(self, signal1: List[float], signal2: List[float]) -> float:
        bark_spectrum1: List[float] = self._compute_bark_spectrum(signal1)
        bark_spectrum2: List[float] = self._compute_bark_spectrum(signal2)

        squared_diffs: List[float] = [(bark_spectrum1[i] - bark_spectrum2[i]) ** 2 for i in range(min(len(bark_spectrum1), len(bark_spectrum2)))]

        mean_squared_diff: float = sum(squared_diffs) / len(squared_diffs)
        return math.sqrt(mean_squared_diff)

import math
from typing import List

class ItakuraSaitoDistance(Distance):

    def __init__(self) -> None:
        super().__init__()

    def _power_spectrum(self, signal: List[float]) -> List[float]:
        N: int = len(signal)
        power_spectrum: List[float] = [0.0] * N

        for i in range(N):
            power_spectrum[i] = signal[i] ** 2
        
        return power_spectrum

    def compute_is_distance(self, signal1: List[float], signal2: List[float]) -> float:
        if len(signal1) != len(signal2):
            raise ValueError("Both signals must have the same length.")

        power_spectrum1: List[float] = self._power_spectrum(signal1)
        power_spectrum2: List[float] = self._power_spectrum(signal2)
        
        is_distance: float = 0.0
        for ps1, ps2 in zip(power_spectrum1, power_spectrum2):
            if ps2 > 0:
                is_distance += (ps1 / ps2) - math.log(ps1 / ps2) - 1
        
        return is_distance
        
import math
from typing import List

class SignalToNoiseRatio(Distance):

    def __init__(self) -> None:
        super().__init__()

    def _power(self, signal: List[float]) -> float:
        power: float = sum(s ** 2 for s in signal) / len(signal)
        return power

    def compute_snr(self, signal: List[float], noise: List[float]) -> float:
        if len(signal) != len(noise):
            raise ValueError("Signal and noise must have the same length.")

        signal_power: float = self._power(signal)
        noise_power: float = self._power(noise)

        if noise_power == 0:
            raise ValueError("Noise power is zero, cannot compute SNR.")

        snr: float = 10 * math.log10(signal_power / noise_power)
        return snr


import math

class PeakSignalToNoiseRatio(Distance):

    def __init__(self) -> None:
        super().__init__()

    def _mean_squared_error(self, signal1: List[float], signal2: List[float]) -> float:
        if len(signal1) != len(signal2):
            raise ValueError("Both signals must have the same length.")

        mse: float = sum((s1 - s2) ** 2 for s1, s2 in zip(signal1, signal2)) / len(signal1)
        return mse

    def compute_psnr(self, signal1: List[float], signal2: List[float], max_signal_value: float) -> float:
        mse: float = self._mean_squared_error(signal1, signal2)
        if mse == 0:
            return float('inf')  # Signals are identical

        psnr: float = 10 * math.log10(max_signal_value ** 2 / mse)
        return psnr
        

class EnergyDistance(Distance):

    def __init__(self) -> None:
        super().__init__()

    def _energy(self, signal: List[float]) -> float:
        energy: float = sum(s ** 2 for s in signal)
        return energy

    def compute_energy_distance(self, signal1: List[float], signal2: List[float]) -> float:
        if len(signal1) != len(signal2):
            raise ValueError("Both signals must have the same length.")

        energy1: float = self._energy(signal1)
        energy2: float = self._energy(signal2)

        energy_distance: float = abs(energy1 - energy2)
        return energy_distance
        
class EnvelopeCorrelation(Distance):

    def __init__(self) -> None:
        super().__init__()

    def _envelope(self, signal: List[float]) -> List[float]:
        # Approximation of the envelope using the absolute value of the signal
        envelope: List[float] = [abs(s) for s in signal]
        return envelope

    def _mean(self, data: List[float]) -> float:
        return sum(data) / len(data)

    def _correlation(self, envelope1: List[float], envelope2: List[float]) -> float:
        mean1: float = self._mean(envelope1)
        mean2: float = self._mean(envelope2)

        numerator: float = sum((e1 - mean1) * (e2 - mean2) for e1, e2 in zip(envelope1, envelope2))
        denominator: float = math.sqrt(sum((e1 - mean1) ** 2 for e1 in envelope1) * sum((e2 - mean2) ** 2 for e2 in envelope2))

        if denominator == 0:
            return 0.0  # No correlation if denominator is zero

        return numerator / denominator

    def compute_envelope_correlation(self, signal1: List[float], signal2: List[float]) -> float:
        if len(signal1) != len(signal2):
            raise ValueError("Both signals must have the same length.")

        envelope1: List[float] = self._envelope(signal1)
        envelope2: List[float] = self._envelope(signal2)

        correlation: float = self._correlation(envelope1, envelope2)
        return correlation
        
class ZeroCrossingRateDistance(Distance):

    def __init__(self) -> None:
        super().__init__()

    def _zero_crossing_rate(self, signal: List[float]) -> float:
        zero_crossings: int = 0
        for i in range(1, len(signal)):
            if (signal[i - 1] > 0 and signal[i] < 0) or (signal[i - 1] < 0 and signal[i] > 0):
                zero_crossings += 1

        zcr: float = zero_crossings / len(signal)
        return zcr

    def compute_zcr_distance(self, signal1: List[float], signal2: List[float]) -> float:
        if len(signal1) != len(signal2):
            raise ValueError("Both signals must have the same length.")

        zcr1: float = self._zero_crossing_rate(signal1)
        zcr2: float = self._zero_crossing_rate(signal2)

        zcr_distance: float = abs(zcr1 - zcr2)
        return zcr_distance
        
class CochleagramDistance(Distance):

    def __init__(self, num_bands: int = 40)-> None:
        super().__init__()

        self.num_bands: int = num_bands

    def _bandpass_filter(self, signal: List[float], band_index: int, total_bands: int) -> List[float]:
        # Simplified bandpass filter approximation
        filtered_signal: List[float] = [0.0] * len(signal)
        band_width: float = 0.5 / total_bands
        center_freq: float = (band_index + 0.5) * band_width
        for i in range(len(signal)):
            filtered_signal[i] = signal[i] * center_freq  # Simplified filter effect
        return filtered_signal

    def _cochleagram(self, signal: List[float]) -> List[List[float]]:
        cochleagram: List[List[float]] = []
        for band in range(self.num_bands):
            band_signal: List[float] = self._bandpass_filter(signal, band, self.num_bands)
            cochleagram.append(band_signal)
        return cochleagram

    def _mean_squared_error(self, cochlea1: List[List[float]], cochlea2: List[List[float]]) -> float:
        if len(cochlea1) != len(cochlea2) or len(cochlea1[0]) != len(cochlea2[0]):
            raise ValueError("Both cochleagrams must have the same dimensions.")

        mse: float = 0.0
        for band1, band2 in zip(cochlea1, cochlea2):
            mse += sum((s1 - s2) ** 2 for s1, s2 in zip(band1, band2)) / len(band1)
        return mse / self.num_bands

    def compute_cochleagram_distance(self, signal1: List[float], signal2: List[float]) -> float:
        if len(signal1) != len(signal2):
            raise ValueError("Both signals must have the same length.")

        cochlea1: List[List[float]] = self._cochleagram(signal1)
        cochlea2: List[List[float]] = self._cochleagram(signal2)

        distance: float = self._mean_squared_error(cochlea1, cochlea2)
        return distance


from typing import List
import math

class ChromagramDistance(Distance):

    def __init__(self, num_bins: int = 12) -> None:
        super().__init__()

        self.num_bins: int = num_bins

    def _frequency_to_bin(self, frequency: float) -> int:
        # Simple mapping of frequency to chroma bin
        if frequency>0:
           bin_index: int = int((12 * math.log2(frequency / 440.0) + 69) % 12)
           return bin_index
        else:
           return 0


    def _compute_chromagram(self, signal: List[float]) -> List[float]:
        chroma: List[float] = [0.0] * self.num_bins
        for sample in signal:
            # Simplified frequency estimation from signal sample (placeholder)
            frequency: float = abs(sample) * 1000.0
            bin_index: int = self._frequency_to_bin(frequency)
            chroma[bin_index] += 1

        # Normalize chromagram
        total_count: float = sum(chroma)
        if total_count > 0:
            chroma = [count / total_count for count in chroma]

        return chroma

    def _mean_squared_error(self, chroma1: List[float], chroma2: List[float]) -> float:
        if len(chroma1) != len(chroma2):
            raise ValueError("Both chromagrams must have the same number of bins.")

        mse: float = sum((c1 - c2) ** 2 for c1, c2 in zip(chroma1, chroma2)) / len(chroma1)
        return mse

    def compute_chromagram_distance(self, signal1: List[float], signal2: List[float]) -> float:
        if len(signal1) != len(signal2):
            raise ValueError("Both signals must have the same length.")

        chroma1: List[float] = self._compute_chromagram(signal1)
        chroma2: List[float] = self._compute_chromagram(signal2)

        distance: float = self._mean_squared_error(chroma1, chroma2)
        return distance

import cmath

class SpectrogramDistance(Distance):

    def __init__(self, window_size: int = 256, overlap: int = 128) -> None:
        super().__init__()

        self.window_size: int = window_size
        self.overlap: int = overlap

    def _dft(self, signal: List[float]) -> List[complex]:
        N: int = len(signal)
        return [sum(signal[n] * cmath.exp(-2j * cmath.pi * k * n / N) for n in range(N)) for k in range(N)]

    def _spectrogram(self, signal: List[float]) -> List[List[float]]:
        step: int = self.window_size - self.overlap
        spectrogram: List[List[float]] = []

        for start in range(0, len(signal) - self.window_size + 1, step):
            windowed_signal: List[float] = signal[start:start + self.window_size]
            dft_result: List[complex] = self._dft(windowed_signal)
            magnitude: List[float] = [abs(freq) for freq in dft_result]
            spectrogram.append(magnitude)

        return spectrogram

    def _mean_squared_error(self, spec1: List[List[float]], spec2: List[List[float]]) -> float:
        if len(spec1) != len(spec2) or len(spec1[0]) != len(spec2[0]):
            raise ValueError("Both spectrograms must have the same dimensions.")

        mse: float = sum(
            (frame1[i] - frame2[i]) ** 2 for frame1, frame2 in zip(spec1, spec2)
            for i in range(len(frame1))
        ) / (len(spec1) * len(spec1[0]))
        return mse

    def compute_spectrogram_distance(self, signal1: List[float], signal2: List[float]) -> float:
        if len(signal1) != len(signal2):
            raise ValueError("Both signals must have the same length.")

        spectrogram1: List[List[float]] = self._spectrogram(signal1)
        spectrogram2: List[List[float]] = self._spectrogram(signal2)

        distance: float = self._mean_squared_error(spectrogram1, spectrogram2)
        return distance


import cmath

class CQTDistance(Distance):

    def __init__(self, num_bins: int = 24, window_size: int = 512) -> None:
        super().__init__()

        self.num_bins: int = num_bins
        self.window_size: int = window_size

    def _dft(self, signal: List[float]) -> List[complex]:
        N: int = len(signal)
        return [sum(signal[n] * cmath.exp(-2j * cmath.pi * k * n / N) for n in range(N)) for k in range(N)]

    def _cqt(self, signal: List[float]) -> List[List[float]]:
        step: int = self.window_size
        cqt_matrix: List[List[float]] = []

        for start in range(0, len(signal) - self.window_size + 1, step):
            windowed_signal: List[float] = signal[start:start + self.window_size]
            dft_result: List[complex] = self._dft(windowed_signal)

            # Compute magnitude and split into bins
            magnitude: List[float] = [abs(freq) for freq in dft_result]
            cqt_bins: List[float] = [sum(magnitude[i] for i in range(self.num_bins))]  # Simplified CQT binning
            cqt_matrix.append(cqt_bins)

        return cqt_matrix

    def _mean_squared_error(self, cqt1: List[List[float]], cqt2: List[List[float]]) -> float:
        if len(cqt1) != len(cqt2) or len(cqt1[0]) != len(cqt2[0]):
            raise ValueError("Both CQT matrices must have the same dimensions.")

        mse: float = sum(
            (frame1[i] - frame2[i]) ** 2
            for frame1, frame2 in zip(cqt1, cqt2)
            for i in range(len(frame1))
        ) / (len(cqt1) * len(cqt1[0]))
        return mse

    def compute_cqt_distance(self, signal1: List[float], signal2: List[float]) -> float:
        if len(signal1) != len(signal2):
            raise ValueError("Both signals must have the same length.")

        cqt1: List[List[float]] = self._cqt(signal1)
        cqt2: List[List[float]] = self._cqt(signal2)

        distance: float = self._mean_squared_error(cqt1, cqt2)
        return distance

import wave
from typing import Tuple

class CepstralDistance(Distance):

    def __init__(self, sample_rate: int = 16000, frame_size: int = 512, num_coefficients: int = 13) -> None:
        """
        Initializes the CepstralDistance class with the specified parameters.
        
        :param sample_rate: The sampling rate of the audio signal.
        :param frame_size: The size of each frame used for analysis.
        :param num_coefficients: The number of cepstral coefficients to extract.
        """
        super().__init__()

        self.sample_rate: int = sample_rate
        self.frame_size: int = frame_size
        self.num_coefficients: int = num_coefficients

    def read_audio(self, filepath: str) -> Tuple[List[float], int]:
        """
        Reads the audio file and returns the waveform data along with the sample rate.
        
        :param filepath: Path to the audio file.
        :return: Tuple containing the audio data (as a list of floats) and the sample rate.
        """
        with wave.open(filepath, 'rb') as wav_file:
            n_frames: int = wav_file.getnframes()
            audio_data: List[float] = list(wav_file.readframes(n_frames))
            return audio_data, wav_file.getframerate()

    def compute_cepstral_coefficients(self, signal: List[float]) -> List[float]:
        """
        Computes the cepstral coefficients of a given audio signal.
        
        :param signal: The input audio signal as a list of floats.
        :return: The cepstral coefficients as a list of floats.
        """
        # Compute the power spectrum (simplified for the example)
        power_spectrum: List[float] = [math.log(abs(s)) for s in signal if s!=0]

        # Apply the inverse Fourier transform to obtain cepstral coefficients
        cepstrum: List[float] = self.inverse_fft(power_spectrum)

        # Return only the first 'num_coefficients' coefficients
        return cepstrum[:self.num_coefficients]

    def inverse_fft(self, spectrum: List[float]) -> List[float]:
        """
        Computes the inverse FFT (simplified for illustration purposes).
        
        :param spectrum: The power spectrum of the audio signal.
        :return: The time-domain cepstral coefficients as a list of floats.
        """
        return [math.exp(s) for s in spectrum]  # Simplified inverse

    def calculate_distance(self, cepstral_1: List[float], cepstral_2: List[float]) -> float:
        """
        Computes the Euclidean distance between two sets of cepstral coefficients.
        
        :param cepstral_1: The first set of cepstral coefficients.
        :param cepstral_2: The second set of cepstral coefficients.
        :return: The cepstral distance as a float.
        """
        return math.sqrt(sum((c1 - c2) ** 2 for c1, c2 in zip(cepstral_1, cepstral_2)))

    def compute_cepstral_distance(self, file1: str, file2: str) -> float:
        """
        Computes the Cepstral Distance between two audio files.
        
        :param file1: Path to the first audio file.
        :param file2: Path to the second audio file.
        :return: The Cepstral Distance as a float.
        """
        audio_data_1, sample_rate_1 = self.read_audio(file1)
        audio_data_2, sample_rate_2 = self.read_audio(file2)

        if sample_rate_1 != sample_rate_2:
            raise ValueError("Sample rates of the two audio files must be the same.")

        cepstral_1: List[float] = self.compute_cepstral_coefficients(audio_data_1)
        cepstral_2: List[float] = self.compute_cepstral_coefficients(audio_data_2)

        distance: float = self.calculate_distance(cepstral_1, cepstral_2)
        return distance

import numpy as np
import wave
import struct

def generate_sine_wave(frequency: float, duration: float, sample_rate: int = 44100, amplitude: float = 0.5) -> np.ndarray:
    """
    Generates a sine wave of a given frequency and duration.
    
    :param frequency: Frequency of the sine wave in Hz.
    :param duration: Duration of the sine wave in seconds.
    :param sample_rate: Sampling rate in Hz.
    :param amplitude: Amplitude of the wave (between 0 and 1).
    :return: NumPy array containing the sine wave samples.
    """
    t = np.linspace(0, duration, int(sample_rate * duration), False)
    wave_data = amplitude * np.sin(2 * np.pi * frequency * t)
    return wave_data

def save_wave(filename: str, data: np.ndarray, sample_rate: int = 44100) -> None:
    """
    Saves the NumPy array as a .wav file.
    
    :param filename: The name of the output .wav file.
    :param data: The audio data to save.
    :param sample_rate: The sampling rate in Hz.
    """
    n_samples = data.shape[0]
    wav_file = wave.open(filename, 'w')
    wav_file.setparams((1, 2, sample_rate, n_samples, 'NONE', 'not compressed'))

    for sample in data:
        wav_file.writeframes(struct.pack('<h', int(sample * 32767)))

    wav_file.close()
'''
les fichier ont été générés

if __name__ == "__main__":
    # Generate two different sine wave signals
    duration: float = 2.0  # seconds

    sine_wave1 = generate_sine_wave(frequency=440.0, duration=duration)  # A4 note (440 Hz)
    sine_wave2 = generate_sine_wave(frequency=523.25, duration=duration)  # C5 note (523.25 Hz)

    # Save the generated sine waves to two .wav files
    save_wave("../sample/audio1.wav", sine_wave1)
    save_wave("../sample/audio2.wav", sine_wave2)

    print("Two audio files 'audio1.wav' and 'audio2.wav' have been generated.")
'''
if __name__ == "__main__":
    # Example usage
    file1: str = "../sample/audio1.wav"
    file2: str = "../sample/audio2.wav"

    cepstral_distance_calculator = CepstralDistance()
    distance: float = cepstral_distance_calculator.compute_cepstral_distance(file1, file2)

    print(f"Cepstral Distance: {distance}")
